# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import print_function
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Module for writing correlator output to a FITS IDI file.  The classes and 
functions defined in this module are based heavily off the lwda_fits library.

.. note::
    Some software that deal with FITS IDI files, e.g. AIPS, does not support
    FITS IDI files with more than 99 antennas.  See the :class:`lsl.writer.fitsidi.aips`
    class for a FITS IDI writer that respects this limit.
    
.. versionchanged:: 1.1.4
    Fixed a conjugation problem in the visibilities saved to a FITS-IDI file

.. versionchanged:: 1.2.2
    Added support for writing multiple IFs to the same FITS-IDI file
"""

import os
import re
import math
import ephem
import numpy
import pyfits
from datetime import datetime

from lsl import astro
from lsl.misc import geodesy
from lsl.common import constants
from lsl.misc.total_sorting import cmp_to_total

try:
    from collections import OrderedDict
except ImportError:
    from lsl.misc.OrderedDict import OrderedDict


__version__ = '0.9'
__revision__ = '$Rev: 2740 $'
__all__ = ['IDI', 'AIPS', 'ExtendedIDI', 'StokesCodes', 'NumericStokes', 
        '__version__', '__revision__', '__all__']


IDIVersion = (3, 0)

StokesCodes = { 'I':  1,  'Q': 2,   'U':  3,  'V':  4, 
            'RR': -1, 'LL': -2, 'RL': -3, 'LR': -4, 
            'XX': -5, 'YY': -6, 'XY': -7, 'YX': -8}

NumericStokes = { 1: 'I',   2: 'Q',   3: 'U',   4: 'V', 
            -1: 'RR', -2: 'LL', -3: 'RL', -4: 'LR', 
            -5: 'XX', -6: 'YY', -7: 'XY', -8: 'YX'}


def mergeBaseline(ant1, ant2, shift=16):
    """
    Merge two stand ID numbers into a single baseline using the specified bit 
    shift size.
    """
    
    return (ant1 << shift) | ant2

def splitBaseline(baseline, shift=16):
    """
    Given a baseline, split it into it consistent stand ID numbers.
    """
    
    part = 2**shift - 1
    return (baseline >> shift) & part, baseline & part



class IDI(object):
    """
    Class for storing visibility data and writing the data, along with array
    geometry, frequency setup, etc., to a FITS IDI file that can be read into 
    AIPS via the FITLD task.
    """
    
    _MAX_ANTS = 255
    _PACKING_BIT_SHIFT = 8
    
    class _Antenna(object):
        """
        Holds information describing the location and properties of an antenna.
        """
        
        def __init__(self, id, x, y, z, bits=8):
            self.id = id
            self.x = x
            self.y = y
            self.z = z
            self.levels = bits
            self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': [0.0, 0.0]}
            self.polB = {'Type': 'Y', 'Angle': 90.0, 'Cal': [0.0, 0.0]}
            
        def getName(self):
            return "LWA%03i" % self.id
            
    class _Frequency:
        """
        Holds information about the frequency setup used in the file.
        """

        def __init__(self, offset, channelWidth, bandwidth):
            self.id = 1
            self.bandFreq = offset
            self.chWidth = channelWidth
            self.totalBW = bandwidth
            self.sideBand = 1
            self.baseBand = 0
            
    @cmp_to_total
    class _UVData(object):
        """
        Represents one UV visibility data set for a given observation time.
        """
    
        def __init__(self, obsTime, intTime, baselines, visibilities, pol=StokesCodes['XX'], source='z'):
            self.obsTime = obsTime
            self.intTime = intTime
            self.baselines = baselines
            self.visibilities = visibilities
            self.pol = pol
            self.source = source
            
        def __cmp__(self, y):
            """
            Function to sort the self.data list in order of time and then 
            polarization code.
            """
            
            sID = (self.obsTime, abs(self.pol))
            yID = (y.obsTime,    abs(y.pol)   )
            
            if sID > yID:
                return 1
            elif sID < yID:
                return -1
            else:
                return 0
                
        def time(self):
            return self.obsTime
            
        def getUVW(self, HA, dec, obs):
            Nbase = len(self.baselines)
            uvw = numpy.zeros((Nbase,3), dtype=numpy.float32)
            
            # Phase center coordinates
            # Convert numbers to radians and, for HA, hours to degrees
            HA2 = HA * 15.0 * numpy.pi/180
            dec2 = dec * numpy.pi/180
            lat2 = obs.lat
            
            # Coordinate transformation matrices
            trans1 = numpy.matrix([[0, -numpy.sin(lat2), numpy.cos(lat2)],
                            [1,  0,               0],
                            [0,  numpy.cos(lat2), numpy.sin(lat2)]])
            trans2 = numpy.matrix([[ numpy.sin(HA2),                  numpy.cos(HA2),                 0],
                            [-numpy.sin(dec2)*numpy.cos(HA2),  numpy.sin(dec2)*numpy.sin(HA2), numpy.cos(dec2)],
                            [ numpy.cos(dec2)*numpy.cos(HA2), -numpy.cos(dec2)*numpy.sin(HA2), numpy.sin(dec2)]])
                    
            for i,(a1,a2) in enumerate(self.baselines):
                # Go from a east, north, up coordinate system to a celestial equation, 
                # east, north celestial pole system
                xyzPrime = a1.stand - a2.stand
                xyz = trans1*numpy.matrix([[xyzPrime[0]],[xyzPrime[1]],[xyzPrime[2]]])
                
                # Go from CE, east, NCP to u, v, w
                temp = trans2*xyz
                uvw[i,:] = numpy.squeeze(temp) / constants.c
                
            return uvw
                
        def argsort(self, mapper=None, shift=16):
            packed = []
            for a1,a2 in self.baselines:
                if mapper is None:
                    s1, s2 = a1.stand.id, a2.stand.id
                else:
                    s1, s2 = mapper[a1.stand.id], mapper[a2.stand.id]
                packed.append( mergeBaseline(s1, s2, shift=shift) )
            packed = numpy.array(packed, dtype=numpy.int32)
            
            return numpy.argsort(packed)
            
    def parseRefTime(self, refTime):
        """
        Given a time as either a integer, float, string, or datetime object, 
        convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'.
        """
        
        # Valid time string (modulo the 'T')
        timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')
        
        if type(refTime) in (int, long, float):
            refDateTime = datetime.utcfromtimestamp(refTime)
            refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == datetime:
            refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
        elif type(refTime) == str:
            # Make sure that the string times are of the correct format
            if re.match(timeRE, refTime) is None:
                raise RuntimeError("Malformed date/time provided: %s" % refTime)
            else:
                refTime = refTime.replace(' ', 'T', 1)
        else:
            raise RuntimeError("Unknown time format provided.")
            
        return refTime
        
    def refTime2AstroDate(self):
        """
        Convert a reference time string to an :class:`lsl.astro.date` object.
        """
        
        dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
        return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))
        
    def __init__(self, filename, refTime=0.0, verbose=False, memmap=None, clobber=False):
        """
        Initialize a new FITS IDI object using a filename and a reference time 
        given in seconds since the UNIX 1970 ephem, a python datetime object, or a 
        string in the format of 'YYYY-MM-DDTHH:MM:SS'.
        
        .. versionchanged:: 1.1.2
            Added the 'memmap' and 'clobber' keywords to control if the file
            is memory mapped and whether or not to overwrite an existing file, 
            respectively.
        """
        
        # File-specific information
        self.filename = filename
        self.verbose = verbose
        
        # Observatory-specific information
        self.siteName = 'Unknown'
        
        # Observation-specific information
        self.refTime = self.parseRefTime(refTime)
        self.nAnt = 0
        self.nChan = 0
        self.nStokes = 0
        self.refVal = 0
        self.refPix = 0
        self.channelWidth = 0
        
        # Parameters that store the meta-data and data
        self.array = []
        self.freq = []
        self.stokes = []
        self.data = []
        
        # Open the file and get going
        if os.path.exists(filename):
            if clobber:
                os.unlink(filename)
            else:
                raise IOError("File '%s' already exists" % filename)
        self.FITS = pyfits.open(filename, mode='append', memmap=memmap)
        
    def setStokes(self, polList):
        """
        Given a list of Stokes parameters, update the object's parameters.
        """
        
        for pol in polList:
            if type(pol) == str:
                numericPol = StokesCodes[pol.upper()]
            else:
                numericPol = pol
                
            if numericPol not in self.stokes:
                self.stokes.append(numericPol)
                
        # Sort into order of 'XX', 'YY', 'XY', and 'YX' or 'I', 'Q', 'U', and 'V'
        self.stokes.sort()
        if self.stokes[0] < 0:
            self.stokes.reverse()
            
        self.nStokes = len(self.stokes)
        
    def setFrequency(self, freq):
        """
        Given a numpy array of frequencies, set the relevant common observation
        parameters and add an entry to the self.freq list.
        """
        
        if self.nChan == 0:
            self.nChan = len(freq)
            self.refVal = freq[0]
            self.refPix = 1
            self.channelWidth = numpy.abs(freq[1] - freq[0])
            offset = 0.0
        else:
            assert(len(freq) == self.nChan)
            offset = freq[0] - self.refVal
        totalWidth = numpy.abs(freq[-1] - freq[0])
        
        freqSetup = self._Frequency(offset, self.channelWidth, totalWidth)
        self.freq.append(freqSetup)
        
    def setGeometry(self, site, antennas, bits=8):
        """
        Given a station and an array of stands, set the relevant common observation
        parameters and add entries to the self.array list.
        
        .. versionchanged:: 0.4.0
            Switched over to passing in Antenna instances generated by the
            :mod:`lsl.common.stations` module instead of a list of stand ID
            numbers.
        """
        
        # Make sure that we have been passed 255 or fewer stands
        if len(antennas) > self._MAX_ANTS:
            raise RuntimeError("FITS IDI supports up to %s antennas only, given %i" % (self._MAX_ANTS, len(antennas)))
            
        # Update the observatory-specific information
        self.siteName = site.name
        
        stands = []
        for ant in antennas:
            stands.append(ant.stand.id)
        stands = numpy.array(stands)
        
        arrayX, arrayY, arrayZ = site.getGeocentricLocation()
        
        xyz = numpy.zeros((len(stands),3))
        for i,ant in enumerate(antennas):
            xyz[i,0] = ant.stand.x
            xyz[i,1] = ant.stand.y
            xyz[i,2] = ant.stand.z
            
        # Create the stand mapper to deal with the fact that stands range from 
        # 1 to 258, not 1 to 255
        mapper = OrderedDict()
        if stands.max() > self._MAX_ANTS:
            enableMapper = True
        else:
            enableMapper = False
            
        ants = []
        topo2eci = site.getECITransform()
        for i in xrange(len(stands)):
            eci = numpy.dot(topo2eci, xyz[i,:])
            ants.append( self._Antenna(stands[i], eci[0], eci[1], eci[2], bits=bits) )
            if enableMapper:
                mapper[stands[i]] = i+1
            else:
                mapper[stands[i]] = stands[i]
                
        # If the mapper has been enabled, tell the user about it
        if enableMapper and self.verbose:
            print("FITS IDI: stand ID mapping enabled")
            for key, value in mapper.iteritems():
                print("FITS IDI:  stand #%i -> mapped #%i" % (key, value))
                
        self.nAnt = len(ants)
        self.array.append( {'center': [arrayX, arrayY, arrayZ], 'ants': ants, 'mapper': mapper, 'enableMapper': enableMapper, 'inputAnts': antennas} )
        
    def addComment(self, comment):
        """
        Add a comment to data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._comments.append( comment )
        except AttributeError:
            self._comments = [comment,]
            
    def addHistory(self, history):
        """
        Add a history entry to the data.
        
        .. versionadded:: 1.2.4
        """
        
        try:
            self._history.append( history )
        except AttributeError:
            self._history = [history,]
            
    def addDataSet(self, obsTime, intTime, baselines, visibilities, pol='XX', source='z'):
        """
        Create a UVData object to store a collection of visibilities.
        
        .. versionchanged:: 0.4.0
            Switched over to passing in Antenna instances generated by the
            :mod:`lsl.common.stations` module instead of a list of stand ID
            as part of the baselines.
            
        .. versionchanged:: 1.1.0
            Added a new 'source' keyword to set the phase center for the data.
            This can either by 'z' for zenith or a ephem.Body instances for a
            point on the sky.
        """
        
        if type(pol) == str:
            numericPol = StokesCodes[pol.upper()]
        else:
            numericPol = pol
            
        self.data.append( self._UVData(obsTime, intTime, baselines, visibilities, pol=numericPol, source=source) )
        
    def write(self):
        """
        Fill in the FITS-IDI file will all of the tables in the 
        correct order.
        """
        
        # Validate
        if self.nStokes == 0:
            raise RuntimeError("No polarization setups defined")
        if len(self.freq) == 0:
            raise RuntimeError("No frequency setups defined")
        if self.nAnt == 0:
            raise RuntimeError("No array geometry defined")
        if len(self.data) == 0:
            raise RuntimeError("No visibility data defined")
            
        # Sort the data set
        self.data.sort()
        
        self._writePrimary()
        self._writeGeometry()
        self._writeFrequency()
        self._writeAntenna()
        self._writeBandpass()
        self._writeSource()
        self._writeData()
        
    def close(self):
        """
        Close out the file.
        """
        
        self.FITS.flush()
        self.FITS.close()
        
    def _addCommonKeywords(self, hdr, name, revision):
        """
        Added keywords common to all table headers.
        """
        
        hdr['EXTNAME'] = (name, 'FITS-IDI table name')
        hdr['EXTVER'] = (1, 'table instance number') 
        hdr['TABREV'] = (revision, 'table format revision number')
        hdr['NO_STKD'] = (self.nStokes, 'number of Stokes parameters')
        hdr['STK_1'] = (self.stokes[0], 'first Stokes parameter')
        hdr['NO_BAND'] = (len(self.freq), 'number of frequency bands')
        hdr['NO_CHAN'] = (self.nChan, 'number of frequency channels')
        hdr['REF_FREQ'] = (self.refVal, 'reference frequency (Hz)')
        hdr['CHAN_BW'] = (self.channelWidth, 'channel bandwidth (Hz)')
        hdr['REF_PIXL'] = (float(self.refPix), 'reference frequency bin')
        
        date = self.refTime.split('-')
        name = "ZA%s%s%s" % (date[0][2:], date[1], date[2])
        hdr['OBSCODE'] = (name, 'zenith all-sky image')
        
        hdr['ARRNAM'] = self.siteName      
        hdr['RDATE'] = (self.refTime, 'file data reference date')
        
    def _writePrimary(self):
        """
        Write the primary HDU to file.
        """
        
        primary = pyfits.PrimaryHDU()
        
        primary.header['NAXIS'] = (0, 'indicates IDI file')
        primary.header['EXTEND'] = (True, 'indicates IDI file')
        primary.header['GROUPS'] = (True, 'indicates IDI file')
        primary.header['GCOUNT'] = 0
        primary.header['PCOUNT'] = 0
        primary.header['OBJECT'] = 'BINARYTB'
        primary.header['TELESCOP'] = self.siteName
        primary.header['INSTRUME'] = self.siteName
        primary.header['OBSERVER'] = ('ZASKY', 'zenith all-sky image')
        primary.header['ORIGIN'] = 'LSL'
        primary.header['CORRELAT'] = ('LWASWC', 'Correlator used')
        primary.header['FXCORVER'] = ('1', 'Correlator version')
        primary.header['LWATYPE'] = ('IDI-ZA', 'LWA FITS file type')
        primary.header['LWAMAJV'] = (IDIVersion[0], 'LWA FITS file format major version')
        primary.header['LWAMINV'] = (IDIVersion[1], 'LWA FITS file format minor version')
        primary.header['DATE-OBS'] = (self.refTime, 'IDI file data collection date')
        ts = str(astro.get_date_from_sys())
        primary.header['DATE-MAP'] = (ts.split()[0], 'IDI file creation date')
        
        # Write the comments and history
        try:
            for comment in self._comments:
                primary.header['comment'] = comment
            del self._comments
        except AttributeError:
            pass
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        try:
            for hist in self._history:
                primary.header['history'] = hist
            del self._history
        except AttributeError:
            pass
            
        self.FITS.append(primary)
        self.FITS.flush()
        
    def _writeGeometry(self):
        """
        Define the Array_Geometry table (group 1, table 1).
        """
        
        names = []
        xyz = numpy.zeros((self.nAnt,3), dtype=numpy.float64)
        for i,ant in enumerate(self.array[0]['ants']):
            xyz[i,0] = ant.x
            xyz[i,1] = ant.y
            xyz[i,2] = ant.z
            names.append(ant.getName())
            
        # Antenna name
        c1 = pyfits.Column(name='ANNAME', format='A8', 
                        array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
        # Station coordinates in meters
        c2 = pyfits.Column(name='STABXYZ', unit='METERS', format='3D', 
                        array=xyz)
        # First order derivative of station coordinates in m/s
        c3 = pyfits.Column(name='DERXYZ', unit='METERS/S', format='3E', 
                        array=numpy.zeros((self.nAnt,3), dtype=numpy.float32))
        # Orbital elements
        c4 = pyfits.Column(name='ORBPARM', format='1D', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
        # Station number
        c5 = pyfits.Column(name='NOSTA', format='1J', 
                        array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
        # Mount type (0 == alt-azimuth)
        c6 = pyfits.Column(name='MNTSTA', format='1J', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.int32))
        # Axis offset in meters
        c7 = pyfits.Column(name='STAXOF', unit='METERS', format='3E', 
                        array=numpy.zeros((self.nAnt,3), dtype=numpy.float32))
                        
        # Define the collection of columns
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7])
        
        # Create the table and fill in the header
        ag = pyfits.new_table(colDefs)
        self._addCommonKeywords(ag.header, 'ARRAY_GEOMETRY', 1)
        
        ag.header['EXTVER'] = (1, 'array ID')
        ag.header['ARRNAM'] = self.siteName
        ag.header['FRAME'] = ('GEOCENTRIC', 'coordinate system')
        ag.header['NUMORB'] = (0, 'number of orbital parameters')
        ag.header['FREQ'] = (self.refVal, 'reference frequency (Hz)')
        ag.header['TIMSYS'] = ('UTC', 'time coordinate system')
        
        date = self.refTime2AstroDate()
        utc0 = date.to_jd()
        gst0 = astro.get_apparent_sidereal_time(utc0)
        ag.header['GSTIA0'] = (gst0 * 15, 'GAST (deg) at RDATE 0 hours')
        
        utc1 = utc0 + 1
        gst1 = astro.get_apparent_sidereal_time(utc1)
        if gst1 < gst0:
            gst1 += 24.0
        ds = gst1 - gst0
        deg = ds * 15.0      
        ag.header['DEGPDY'] = (360.0 + deg, 'rotation rate of the earth (deg/day)')
        
        refDate = self.refTime2AstroDate()
        refMJD = refDate.to_jd() - astro.MJD_OFFSET
        eop = geodesy.getEOP(refMJD)
        if eop is None:
            eop = geodesy.EOP(mjd=refMJD)
            
        ag.header['UT1UTC'] = (eop.utDiff, 'difference UT1 - UTC for reference date')
        ag.header['IATUTC'] = (astro.leap_secs(utc0), 'TAI - UTC for reference date')
        ag.header['POLARX'] = eop.x
        ag.header['POLARY'] = eop.y
        
        ag.header['ARRAYX'] = (self.array[0]['center'][0], 'array ECI X coordinate (m)')
        ag.header['ARRAYY'] = (self.array[0]['center'][1], 'array ECI Y coordinate (m)')
        ag.header['ARRAYZ'] = (self.array[0]['center'][2], 'array ECI Z coordinate (m)')
        
        ag.header['NOSTAMAP'] = (int(self.array[0]['enableMapper']), 'Mapping enabled for stand numbers')
        
        ag.name = 'ARRAY_GEOMETRY'
        self.FITS.append(ag)
        self.FITS.flush()
        
        if self.array[0]['enableMapper']:
            self._writeMapper()
            
    def _writeFrequency(self):
        """
        Define the Frequency table (group 1, table 3).
        """
        
        nBand = len(self.freq)
        
        # Frequency setup number
        c1 = pyfits.Column(name='FREQID', format='1J', 
                        array=numpy.array([self.freq[0].id,], dtype=numpy.int32))
        # Frequency offsets in Hz
        c2 = pyfits.Column(name='BANDFREQ', format='%iD' % nBand, unit='HZ', 
                        array=numpy.array([f.bandFreq for f in self.freq], dtype=numpy.float64).reshape(1,nBand))
        # Channel width in Hz
        c3 = pyfits.Column(name='CH_WIDTH', format='%iE' % nBand, unit='HZ', 
                        array=numpy.array([f.chWidth for f in self.freq], dtype=numpy.float32).reshape(1,nBand))
        # Total bandwidths of bands
        c4 = pyfits.Column(name='TOTAL_BANDWIDTH', format='%iE' % nBand, unit='HZ', 
                        array=numpy.array([f.totalBW for f in self.freq], dtype=numpy.float32).reshape(1,nBand))
        # Sideband flag
        c5 = pyfits.Column(name='SIDEBAND', format='%iJ' % nBand, 
                        array=numpy.array([f.sideBand for f in self.freq], dtype=numpy.int32).reshape(1,nBand))
        # Baseband channel
        c6 = pyfits.Column(name='BB_CHAN', format='%iJ' % nBand, 
                        array=numpy.array([f.baseBand for f in self.freq], dtype=numpy.int32).reshape(1,nBand))
                        
        # Define the collection of columns
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6])
        
        # Create the table and header
        fq = pyfits.new_table(colDefs)
        self._addCommonKeywords(fq.header, 'FREQUENCY', 1)
        
        # Add the table to the file
        fq.name = 'FREQUENCY'
        self.FITS.append(fq)
        self.FITS.flush()
        
    def _writeAntenna(self):
        """
        Define the Antenna table (group 2, table 1).
        """
        
        nBand = len(self.freq)
        
        # Central time of period covered by record in days
        c1 = pyfits.Column(name='TIME', unit='DAYS', format='1D', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
        # Duration of period covered by record in days
        c2 = pyfits.Column(name='TIME_INTERVAL', unit='DAYS', format='1E', 
                        array=(2*numpy.ones((self.nAnt,), dtype=numpy.float32)))
        # Antenna name
        c3 = pyfits.Column(name='ANNAME', format='A8', 
                        array=self.FITS['ARRAY_GEOMETRY'].data.field('ANNAME'))
        # Antenna number
        c4 = pyfits.Column(name='ANTENNA_NO', format='1J', 
                        array=self.FITS['ARRAY_GEOMETRY'].data.field('NOSTA'))
        # Array number
        c5 = pyfits.Column(name='ARRAY', format='1J', 
                        array=numpy.ones((self.nAnt,), dtype=numpy.int32))
        # Frequency setup number
        c6 = pyfits.Column(name='FREQID', format='1J', 
                        array=(numpy.zeros((self.nAnt,), dtype=numpy.int32) + self.freq[0].id))
        # Number of digitizer levels
        c7 = pyfits.Column(name='NO_LEVELS', format='1J', 
                        array=numpy.array([ant.levels for ant in self.array[0]['ants']]))
        # Feed A polarization label
        c8 = pyfits.Column(name='POLTYA', format='A1', 
                        array=numpy.array([ant.polA['Type'] for ant in self.array[0]['ants']]))
        # Feed A orientation in degrees
        c9 = pyfits.Column(name='POLAA', format='%iE' % nBand,  unit='DEGREES', 
                        array=numpy.array([[ant.polA['Angle'],]*nBand for ant in self.array[0]['ants']], dtype=numpy.float32))
        # Feed A polarization parameters
        c10 = pyfits.Column(name='POLCALA', format='%iE' % (2*nBand), 
                        array=numpy.concatenate([[ant.polA['Cal'],]*nBand for ant in self.array[0]['ants']]).astype(numpy.float32))
        # Feed B polarization label
        c11 = pyfits.Column(name='POLTYB', format='A1', 
                        array=numpy.array([ant.polB['Type'] for ant in self.array[0]['ants']]))
        # Feed B orientation in degrees
        c12 = pyfits.Column(name='POLAB', format='%iE' % nBand,  unit='DEGREES', 
                        array=numpy.array([[ant.polB['Angle'],]*nBand for ant in self.array[0]['ants']], dtype=numpy.float32))
        # Feed B polarization parameters
        c13 = pyfits.Column(name='POLCALB', format='%iE' % (2*nBand), 
                        array=numpy.concatenate([[ant.polB['Cal'],]*nBand for ant in self.array[0]['ants']]).astype(numpy.float32))
                        
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
                            c11, c12, c13])
                            
        # Create the Antenna table and update it's header
        an = pyfits.new_table(colDefs)
        self._addCommonKeywords(an.header, 'ANTENNA', 1)
        
        an.header['NOPCAL'] = (2, 'number of polarization parameters')
        an.header['POLTYPE'] = ('X-Y LIN', 'polarization parameterization')
        
        an.name = 'ANTENNA'
        self.FITS.append(an)
        self.FITS.flush()
        
    def _writeBandpass(self):
        """
        Define the Bandpass table (group 2, table 3).
        """
        
        nBand = len(self.freq)
        
        # Central time of period covered by record in days
        c1 = pyfits.Column(name='TIME', unit='DAYS', format='1D', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.float64))
        # Duration of period covered by record in days
        c2 = pyfits.Column(name='TIME_INTERVAL', unit='DAYS', format='1E',
                        array=(2*numpy.ones((self.nAnt,), dtype=numpy.float32)))
        # Source ID
        c3 = pyfits.Column(name='SOURCE_ID', format='1J', 
                        array=numpy.zeros((self.nAnt,), dtype=numpy.int32))
        # Antenna number
        c4 = pyfits.Column(name='ANTENNA_NO', format='1J', 
                        array=self.FITS['ANTENNA'].data.field('ANTENNA_NO'))
        # Array number
        c5 = pyfits.Column(name='ARRAY', format='1J', 
                        array=numpy.ones((self.nAnt,), dtype=numpy.int32))
        # Frequency setup number
        c6 = pyfits.Column(name='FREQID', format='1J',
                        array=(numpy.zeros((self.nAnt,), dtype=numpy.int32) + self.freq[0].id))
        # Bandwidth in Hz
        c7 = pyfits.Column(name='BANDWIDTH', unit='HZ', format='1E',
                        array=(numpy.zeros((self.nAnt,), dtype=numpy.float32)+self.freq[0].totalBW))
        # Band frequency in Hz
        c8 = pyfits.Column(name='BAND_FREQ', unit='HZ', format='%iD' % nBand,
                        array=(numpy.zeros((self.nAnt,nBand,), dtype=numpy.float64)+[f.bandFreq for f in self.freq]))
        # Reference antenna number (pol. 1)
        c9 = pyfits.Column(name='REFANT_1', format='1J',
                        array=numpy.ones((self.nAnt,), dtype=numpy.int32))
        # Real part of the bandpass (pol. 1)
        c10 = pyfits.Column(name='BREAL_1', format='%iE' % (self.nChan*nBand),
                        array=numpy.ones((self.nAnt,nBand*self.nChan), dtype=numpy.float32))
        # Imaginary part of the bandpass (pol. 1)
        c11 = pyfits.Column(name='BIMAG_1', format='%iE' % (self.nChan*nBand),
                        array=numpy.zeros((self.nAnt,nBand*self.nChan), dtype=numpy.float32))
        # Reference antenna number (pol. 2)
        c12 = pyfits.Column(name='REFANT_2', format='1J',
                        array=numpy.ones((self.nAnt,), dtype=numpy.int32))
        # Real part of the bandpass (pol. 2)
        c13 = pyfits.Column(name='BREAL_2', format='%iE' % (self.nChan*nBand),
                        array=numpy.ones((self.nAnt,nBand*self.nChan), dtype=numpy.float32))
        # Imaginary part of the bandpass (pol. 2)
        c14 = pyfits.Column(name='BIMAG_2', format='%iE' % (self.nChan*nBand),
                        array=numpy.zeros((self.nAnt,nBand*self.nChan), dtype=numpy.float32))
                        
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
                            c11, c12, c13, c14])
                            
        # Create the Bandpass table and update its header
        bp = pyfits.new_table(colDefs)
        self._addCommonKeywords(bp.header, 'BANDPASS', 1)
        
        bp.header['NO_ANT'] = self.nAnt
        bp.header['NO_POL'] = 2
        bp.header['NO_BACH'] = self.nChan
        bp.header['STRT_CHN'] = self.refPix
        
        bp.name = 'BANDPASS'
        self.FITS.append(bp)
        self.FITS.flush()
        
    def _writeSource(self):
        """
        Define the Source table (group 1, table 2).
        """
        
        nBand = len(self.freq)
        
        (arrPos, ag) = self.readArrayGeometry()
        ids = ag.keys()
        
        obs = ephem.Observer()
        obs.lat = arrPos.lat * numpy.pi/180
        obs.lon = arrPos.lng * numpy.pi/180
        obs.elev = arrPos.elv * numpy.pi/180
        obs.pressure = 0
        
        nameList = []
        raList = []
        decList = []
        raPoList = []
        decPoList = []
        sourceID = 0
        for dataSet in self.data:
            if dataSet.pol == self.stokes[0]:
                utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                date = astro.get_date(utc)
                date.hours = 0
                date.minutes = 0
                date.seconds = 0
                utc0 = date.to_jd()
                
                obs.date = utc - astro.DJD_OFFSET
                
                try:
                    currSourceName = dataSet.source.name
                except AttributeError:
                    currSourceName = dataSet.source
                
                if currSourceName not in nameList:
                    sourceID += 1
                    
                    if dataSet.source == 'z':
                        ## Zenith pointings
                        equ = astro.equ_posn( obs.sidereal_time()*180/numpy.pi, obs.lat*180/numpy.pi )
                        
                        # format 'source' name based on local sidereal time
                        raHms = astro.deg_to_hms(equ.ra)
                        (tsecs, secs) = math.modf(raHms.seconds)
                        
                        name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
                        equPo = astro.get_equ_prec2(equ, utc, astro.J2000_UTC_JD)
                        
                    else:
                        ## Real-live sources (ephem.Body instances)
                        name = dataSet.source.name
                        equ = astro.equ_posn(dataSet.source.ra*180/numpy.pi, dataSet.source.dec*180/numpy.pi)
                        equPo = astro.equ_posn(dataSet.source.a_ra*180/numpy.pi, dataSet.source.a_dec*180/numpy.pi)
                        
                    # current apparent zenith equatorial coordinates
                    raList.append(equ.ra)
                    decList.append(equ.dec)
                    
                    # J2000 zenith equatorial coordinates
                    raPoList.append(equPo.ra)
                    decPoList.append(equPo.dec)
                    
                    # name
                    nameList.append(name)
                    
        nSource = len(nameList)
        
        # Save these for later since we might need them
        self._sourceTable = nameList
        
        # Source ID number
        c1 = pyfits.Column(name='SOURCE_ID', format='1J', 
                        array=numpy.arange(1, nSource+1, dtype=numpy.int32))
        # Source name
        c2 = pyfits.Column(name='SOURCE', format='A16', 
                        array=numpy.array(nameList))
        # Source qualifier
        c3 = pyfits.Column(name='QUAL', format='1J', 
                        array=numpy.zeros((nSource,), dtype=numpy.int32))
        # Calibrator code
        c4 = pyfits.Column(name='CALCODE', format='A4', 
                        array=numpy.array(('   ',)).repeat(nSource))
        # Frequency group ID
        c5 = pyfits.Column(name='FREQID', format='1J', 
                        array=(numpy.zeros((nSource,), dtype=numpy.int32)+self.freq[0].id))
        # Stokes I flux density in Jy
        c6 = pyfits.Column(name='IFLUX', format='%iE' % nBand, unit='JY', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Stokes I flux density in Jy
        c7 = pyfits.Column(name='QFLUX', format='%iE' % nBand, unit='JY', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Stokes I flux density in Jy
        c8 = pyfits.Column(name='UFLUX', format='%iE' % nBand, unit='JY', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Stokes I flux density in Jy
        c9 = pyfits.Column(name='VFLUX', format='%iE' % nBand, unit='JY', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Spectral index
        c10 = pyfits.Column(name='ALPHA', format='%iE' % nBand, 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Frequency offset in Hz
        c11 = pyfits.Column(name='FREQOFF', format='%iE' % nBand, unit='HZ', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float32))
        # Mean equinox and epoch
        c12 = pyfits.Column(name='EQUINOX', format='A8',
                        array=numpy.array(('J2000',)).repeat(nSource))
        c13 = pyfits.Column(name='EPOCH', format='1D', unit='YEARS', 
                        array=numpy.zeros((nSource,), dtype=numpy.float64) + 2000.0)
        # Apparent right ascension in degrees
        c14 = pyfits.Column(name='RAAPP', format='1D', unit='DEGREES', 
                        array=numpy.array(raList))
        # Apparent declination in degrees
        c15 = pyfits.Column(name='DECAPP', format='1D', unit='DEGREES', 
                        array=numpy.array(decList))
        # Right ascension at mean equinox in degrees
        c16 = pyfits.Column(name='RAEPO', format='1D', unit='DEGREES', 
                        array=numpy.array(raPoList))
        # Declination at mean equinox in degrees
        c17 = pyfits.Column(name='DECEPO', format='1D', unit='DEGREES', 
                        array=numpy.array(decPoList))
        # Systemic velocity in m/s
        c18 = pyfits.Column(name='SYSVEL', format='%iD' % nBand, unit='M/SEC', 
                        array=numpy.zeros((nSource,nBand), dtype=numpy.float64))
        # Velocity type
        c19 = pyfits.Column(name='VELTYP', format='A8', 
                        array=numpy.array(('GEOCENTR',)).repeat(nSource))
        # Velocity definition
        c20 = pyfits.Column(name='VELDEF', format='A8', 
                        array=numpy.array(('OPTICAL',)).repeat(nSource))
        # Line rest frequency in Hz
        c21 = pyfits.Column(name='RESTFREQ', format='%iD' % nBand, unit='HZ', 
                        array=(numpy.zeros((nSource,nBand), dtype=numpy.float64) + [f.bandFreq+self.refVal for f in self.freq]))
        # Proper motion in RA in degrees/day
        c22 = pyfits.Column(name='PMRA', format='1D', unit='DEG/DAY', 
                        array=numpy.zeros((nSource,), dtype=numpy.float64))
        # Proper motion in Dec in degrees/day
        c23 = pyfits.Column(name='PMDEC', format='1D', unit='DEG/DAY', 
                        array=numpy.zeros((nSource,), dtype=numpy.float64))
        # Parallax of source in arc sec.
        c24 = pyfits.Column(name='PARALLAX', format='1E', unit='ARCSEC', 
                        array=numpy.zeros((nSource,), dtype=numpy.float32))
                        
        # Define the collection of columns
        colDefs = pyfits.ColDefs([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, 
                            c11, c12, c13, c14, c15, c16, c17, c18, c19, c20, 
                            c21, c22, c23, c24])
                            
        # Create the Source table and update its header
        sr = pyfits.new_table(colDefs)
        self._addCommonKeywords(sr.header, 'SOURCE', 1)
        
        sr.name = 'SOURCE'
        self.FITS.append(sr)
        self.FITS.flush()
        
    def _writeData(self):
        """
        Define the UV_Data table (group 3, table 1).
        """
        
        nBand = len(self.freq)
        
        (arrPos, ag) = self.readArrayGeometry()
        (mapper, inverseMapper) = self.readArrayMapper()
        ids = ag.keys()
        
        obs = ephem.Observer()
        obs.lat = arrPos.lat * numpy.pi/180
        obs.lon = arrPos.lng * numpy.pi/180
        obs.elev = arrPos.elv * numpy.pi/180
        obs.pressure = 0
        
        mList = []
        uList = []
        vList = []
        wList = []
        timeList = []
        dateList = []
        intTimeList = []
        blineList = []
        nameList = []
        sourceList = []
        while True:
            # Get the next data set to process
            try:
                dataSet = self.data.pop(0)
            except IndexError:
                break
                
            # Sort the data by packed baseline
            try:
                order
            except NameError:
                order = dataSet.argsort(mapper=mapper, shift=self._PACKING_BIT_SHIFT)
                
            # Deal with defininig the values of the new data set
            if dataSet.pol == self.stokes[0]:
                ## Figure out the new date/time for the observation
                utc = astro.taimjd_to_utcjd(dataSet.obsTime)
                date = astro.get_date(utc)
                date.hours = 0
                date.minutes = 0
                date.seconds = 0
                utc0 = date.to_jd()
                
                ## Update the observer so we can figure out where the source is
                obs.date = utc - astro.DJD_OFFSET
                if dataSet.source == 'z':
                    ### Zenith pointings
                    equ = astro.equ_posn( obs.sidereal_time()*180/numpy.pi, obs.lat*180/numpy.pi )
                    
                    ### format 'source' name based on local sidereal time
                    raHms = astro.deg_to_hms(equ.ra)
                    (tsecs, secs) = math.modf(raHms.seconds)
                    name = "ZA%02d%02d%02d%01d" % (raHms.hours, raHms.minutes, int(secs), int(tsecs * 10.0))
                else:
                    ### Real-live sources (ephem.Body instances)
                    name = dataSet.source.name
                    
                ## Update the source ID
                sourceID = self._sourceTable.index(name) + 1
                
                ## Compute the uvw coordinates of all baselines
                if dataSet.source == 'z':
                    HA = 0.0
                    dec = equ.dec
                else:
                    HA = (obs.sidereal_time() - dataSet.source.ra) * 12/numpy.pi
                    dec = dataSet.source.dec * 180/numpy.pi
                uvwCoords = dataSet.getUVW(HA, dec, obs)
                
                ## Populate the metadata
                ### Add in the new baselines
                try:
                    blineList.extend( baselineMapped )
                except NameError:
                    baselineMapped = []
                    for o in order:
                        antenna1, antenna2 = dataSet.baselines[o]
                        if mapper is None:
                            stand1, stand2 = antenna1.stand.id, antenna2.stand.id
                        else:
                            stand1, stand2 = mapper[antenna1.stand.id], mapper[antenna2.stand.id]
                        baselineMapped.append( mergeBaseline(stand1, stand2, shift=self._PACKING_BIT_SHIFT) ) 
                    blineList.extend( baselineMapped )
                    
                ### Add in the new u, v, and w coordinates
                uList.extend( uvwCoords[order,0] )
                vList.extend( uvwCoords[order,1] )
                wList.extend( uvwCoords[order,2] )
                
                ### Add in the new date/time and integration time
                dateList.extend( [utc0 for bl in dataSet.baselines] )
                timeList.extend( [utc-utc0 for bl in dataSet.baselines] )
                intTimeList.extend( [dataSet.intTime for bl in dataSet.baselines] )
                
                ### Add in the new new source ID and name
                sourceList.extend( [sourceID for bl in dataSet.baselines] )
                nameList.extend( [name for bl in dataSet.baselines] )
                
                ### Zero out the visibility data
                try:
                    matrix *= 0.0
                except NameError:
                    matrix = numpy.zeros((len(order), self.nStokes*self.nChan*nBand), dtype=numpy.complex64)
                    
            # Save the visibility data in the right order
            # NOTE:  This is this conjugate since there seems to be a convention mis-match
            #        between LSL and AIPS/the FITS-IDI convention.
            matrix[:,self.stokes.index(dataSet.pol)::self.nStokes] = dataSet.visibilities[order,:].conj()
            
            # Deal with saving the data once all of the polarizations have been added to 'matrix'
            if dataSet.pol == self.stokes[-1]:
                mList.append( matrix.view(numpy.float32)*1.0 )
                
            # Cleanup
            del dataSet
            
        nBaseline = len(blineList)
        nSource = len(nameList)
        
        # Visibility Data
        c1 = pyfits.Column(name='FLUX', format='%iE' % (2*self.nStokes*self.nChan*nBand), unit='UNCALIB', 
                        array=numpy.concatenate(mList))
        # Baseline number (first*256+second)
        c2 = pyfits.Column(name='BASELINE', format='1J', 
                        array=numpy.array(blineList))
        # Julian date at 0h
        c3 = pyfits.Column(name='DATE', format='1D', unit='DAYS',
                        array = numpy.array(dateList))
        # Time elapsed since 0h
        c4 = pyfits.Column(name='TIME', format='1D', unit = 'DAYS', 
                        array = numpy.array(timeList))
        # Integration time (seconds)
        c5 = pyfits.Column(name='INTTIM', format='1D', unit='SECONDS', 
                        array=numpy.array(intTimeList, dtype=numpy.float32))
        # U coordinate (light seconds)
        c6 = pyfits.Column(name='UU', format='1E', unit='SECONDS', 
                        array=numpy.array(uList, dtype=numpy.float32))
        # V coordinate (light seconds)
        c7 = pyfits.Column(name='VV', format='1E', unit='SECONDS', 
                        array=numpy.array(vList, dtype=numpy.float32))
        # W coordinate (light seconds)
        c8 = pyfits.Column(name='WW', format='1E', unit='SECONDS', 
                        array=numpy.array(wList, dtype=numpy.float32))
        # Source ID number
        c9 = pyfits.Column(name='SOURCE', format='1J', 
                        array=numpy.array(sourceList))
        # Frequency setup number
        c10 = pyfits.Column(name='FREQID', format='1J', 
                        array=(numpy.zeros((nBaseline,), dtype=numpy.int32) + self.freq[0].id))
        # Filter number
        c11 = pyfits.Column(name='FILTER', format='1J', 
                        array=numpy.zeros((nBaseline,), dtype=numpy.int32))
        # Gate ID number
        c12 = pyfits.Column(name='GATEID', format='1J', 
                        array=numpy.zeros((nBaseline,), dtype=numpy.int32))
        # Weights
        c13 = pyfits.Column(name='WEIGHT', format='%iE' % (self.nStokes*self.nChan*nBand), 
                        array=numpy.ones((nBaseline, self.nStokes*self.nChan*nBand), dtype=numpy.float32))
                        
        colDefs = pyfits.ColDefs([c6, c7, c8, c3, c4, c2, c11, c9, c10, c5, 
                        c13, c12, c1])
                        
        # Create the UV Data table and update its header
        uv = pyfits.new_table(colDefs)
        self._addCommonKeywords(uv.header, 'UV_DATA', 1)
        
        uv.header['NMATRIX'] = (1, 'number of UV data matricies')
        uv.header['MAXIS'] = (6, 'number of UV data matrix axes')
        uv.header['TMATX13'] = (True, 'axis 13 contains UV matrix')
        
        uv.header['MAXIS1'] = (2, 'number of pixels in COMPLEX axis')
        uv.header['CTYPE1'] = ('COMPLEX', 'axis 1 is COMPLEX axis')
        uv.header['CDELT1'] = 1.0
        uv.header['CRPIX1'] = 1.0
        uv.header['CRVAL1'] = 1.0
        
        uv.header['MAXIS2'] = (self.nStokes, 'number of pixels in STOKES axis')
        uv.header['CTYPE2'] = ('STOKES', 'axis 2 is STOKES axis (polarization)')
        if self.stokes[0] < 0:
            uv.header['CDELT2'] = -1.0
        else:
            uv.header['CDELT2'] = 1.0
        uv.header['CRPIX2'] = 1.0
        uv.header['CRVAL2'] = float(self.stokes[0])
        
        uv.header['MAXIS3'] = (self.nChan, 'number of pixels in FREQ axis')
        uv.header['CTYPE3'] = ('FREQ', 'axis 3 is FREQ axis (frequency)')
        uv.header['CDELT3'] = self.freq[0].chWidth
        uv.header['CRPIX3'] = self.refPix
        uv.header['CRVAL3'] = self.refVal
        
        uv.header['MAXIS4'] = (nBand, 'number of pixels in BAND (IF) axis')
        uv.header['CTYPE4'] = ('BAND', 'axis 4 is BAND axis')
        uv.header['CDELT4'] = 1.0
        uv.header['CRPIX4'] = 1.0
        uv.header['CRVAL4'] = 1.0
        
        uv.header['MAXIS5'] = (1, 'number of pixels in RA axis')
        uv.header['CTYPE5'] = ('RA', 'axis 5 is RA axis (position of phase center)')
        uv.header['CDELT5'] = 0.0
        uv.header['CRPIX5'] = 1.0
        uv.header['CRVAL5'] = 0.0
        
        uv.header['MAXIS6'] = (1, 'number of pixels in DEC axis')
        uv.header['CTYPE6'] = ('DEC', 'axis 6 is DEC axis (position of phase center)')
        uv.header['CDELT6'] = 0.0
        uv.header['CRPIX6'] = 1.0
        uv.header['CRVAL6'] = 0.0
        
        uv.header['TELESCOP'] = self.siteName
        uv.header['OBSERVER'] = 'ZASKY'
        uv.header['SORT'] = ('TB', 'data is sorted in [time,baseline] order')
        
        uv.header['VISSCALE'] = (1.0, 'UV data scale factor')
        
        uv.name = 'UV_DATA'
        self.FITS.append(uv)
        self.FITS.flush()
        
    def _writeMapper(self):
        """
        Write a fits table that contains information about mapping stations 
        numbers to actual antenna numbers.  This information can be backed out of
        the names, but this makes the extraction more programmatic.
        """
        
        c1 = pyfits.Column(name='ANNAME', format='A8', 
                        array=numpy.array([ant.getName() for ant in self.array[0]['ants']]))
        c2 = pyfits.Column(name='NOSTA', format='1J', 
                        array=numpy.array([self.array[0]['mapper'][ant.id] for ant in self.array[0]['ants']]))
        c3 = pyfits.Column(name='NOACT', format='1J', 
                        array=numpy.array([ant.id for ant in self.array[0]['ants']]))
                        
        colDefs = pyfits.ColDefs([c1, c2, c3])
        
        # Create the ID mapping table and update its header
        nsm = pyfits.new_table(colDefs)
        self._addCommonKeywords(nsm.header, 'NOSTA_MAPPER', 1)
        
        nsm.name = 'NOSTA_MAPPER'
        self.FITS.append(nsm)
        self.FITS.flush()
        
    def readArrayGeometry(self):
        """
        Return a tuple with the array geodetic position and the local 
        positions for all antennas defined in the ARRAY_GEOMETRY table.
        """
        
        try:
            ag = self.FITS['ARRAY_GEOMETRY']
        except IndexError:
            raise RuntimeError("File does not have an 'ARRAY_GEOMTRY' table.")
            
        # Array position
        arrayGeo = astro.rect_posn(ag.header['ARRAYX'], ag.header['ARRAYY'], ag.header['ARRAYZ'])
        arrayGeo = astro.get_geo_from_rect(arrayGeo)
        
        # Antenna positions
        antennaGeo = OrderedDict()
        antenna = ag.data
        antennaID = antenna.field('NOSTA')
        antennaPos = iter(antenna.field('STABXYZ'))
        for id in antennaID:
            antennaGeo[id] = next(antennaPos)
            
        # Return
        return (arrayGeo, antennaGeo)
        
    def readArrayMapper(self):
        """
        Return a tuple with the array NOSTA mapper and inverse mapper (both
        dictionaries.  If the stand IDs have not been mapped, return None for
        both.
        """
        
        try:
            nsm = self.FITS['NOSTA_MAPPER']
        except KeyError:
            return (None, None)
            
        # Build the mapper and inverseMapper
        mapper = OrderedDict()
        inverseMapper = OrderedDict()
        nosta = nsm.data.field('NOSTA')
        noact = nsm.data.field('NOACT')
        for idMapped, idActual in zip(nosta, noact):
            mapper[idActual] = idMapped
            inverseMapper[idMapped] = idActual
            
        # Return
        return (mapper, inverseMapper)


class AIPS(IDI):
    """
    Sub-class of the FITS IDI writer for making files that *should* work 
    with AIPS nicely.  AIPS imposes a limit on antenna number of two digits
    (1-99).  This sub-class overwrite the setGeometry() function with one that
    enforces the two digit limit and maps accordingly.  It also sets the FITS
    `LWATYPE` keyword in the primary HDU to a value of `IDI-AIPS-ZA` to 
    distinguish files written by this writer from the standard IDI writer.
    """
    
    _MAX_ANTS = 99
    _PACKING_BIT_SHIFT = 8
    
    class _Antenna(object):
        """
        Holds information describing the location and properties of an antenna.
        """

        def __init__(self, id, x, y, z, bits=8):
            self.id = id
            self.x = x
            self.y = y
            self.z = z
            self.levels = bits
            self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': [0.0, 0.0]}
            self.polB = {'Type': 'Y', 'Angle': 90.0, 'Cal': [0.0, 0.0]}
            
        def getName(self):
            return "L%03i" % self.id
            
    def _writePrimary(self):
        """
        Write the primary HDU to file.
        """
        
        primary = pyfits.PrimaryHDU()
        
        primary.header['NAXIS'] = (0, 'indicates IDI file')
        primary.header['EXTEND'] = (True, 'indicates IDI file')
        primary.header['GROUPS'] = (True, 'indicates IDI file')
        primary.header['GCOUNT'] = 0
        primary.header['PCOUNT'] = 0
        primary.header['OBJECT'] = 'BINARYTB'
        primary.header['TELESCOP'] = self.siteName
        primary.header['INSTRUME'] = self.siteName
        primary.header['OBSERVER'] = ('ZASKY', 'zenith all-sky image')
        primary.header['ORIGIN'] = 'LSL'
        primary.header['CORRELAT'] = ('LWASWC', 'Correlator used')
        primary.header['FXCORVER'] = ('1', 'Correlator version')
        primary.header['LWATYPE'] = ('IDI-AIPS-ZA', 'LWA FITS file type')
        primary.header['LWAMAJV'] = (IDIVersion[0], 'LWA FITS file format major version')
        primary.header['LWAMINV'] = (IDIVersion[1], 'LWA FITS file format minor version')
        primary.header['DATE-OBS'] = (self.refTime, 'IDI file data collection date')
        ts = str(astro.get_date_from_sys())
        primary.header['DATE-MAP'] = (ts.split()[0], 'IDI file creation date')
        
        # Write the comments and history
        try:
            for comment in self._comments:
                primary.header['comment'] = comment
            del self._comments
        except AttributeError:
            pass
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        try:
            for hist in self._history:
                primary.header['history'] = hist
            del self._history
        except AttributeError:
            pass
            
        self.FITS.append(primary)
        self.FITS.flush()


class ExtendedIDI(IDI):
    """
    Sub-class of the FITS IDI writer for making files that support up to
    65,535 antennas.  This is done by changing the packing of baselines 
    stored in the UVDATA table.  The new packing for baseline (i,j) is:
    (i << 16) & (j)
    It also sets the FITS `LWATYPE` keyword in the primary HDU to a value 
    of `IDI-EXTENDED-ZA` to distinguish files written by this writer 
    from the standard IDI writer.
    """
    
    _MAX_ANTS = 65535
    _PACKING_BIT_SHIFT = 16
    
    class _Antenna(object):
        """
        Holds information describing the location and properties of an antenna.
        """
        
        def __init__(self, id, x, y, z, bits=8):
            self.id = id
            self.x = x
            self.y = y
            self.z = z
            self.levels = bits
            self.polA = {'Type': 'X', 'Angle': 0.0, 'Cal': [0.0, 0.0]}
            self.polB = {'Type': 'Y', 'Angle': 90.0, 'Cal': [0.0, 0.0]}
            
        def getName(self):
            return "LWA%05i" % self.id
            
    def _writePrimary(self):
        """
        Write the primary HDU to file.
        """
        
        primary = pyfits.PrimaryHDU()
        
        primary.header['NAXIS'] = (0, 'indicates IDI file')
        primary.header['EXTEND'] = (True, 'indicates IDI file')
        primary.header['GROUPS'] = (True, 'indicates IDI file')
        primary.header['GCOUNT'] = 0
        primary.header['PCOUNT'] = 0
        primary.header['OBJECT'] = 'BINARYTB'
        primary.header['TELESCOP'] = self.siteName
        primary.header['INSTRUME'] = self.siteName
        primary.header['OBSERVER'] = ('ZASKY', 'zenith all-sky image')
        primary.header['ORIGIN'] = 'LSL'
        primary.header['CORRELAT'] = ('LWASWC', 'Correlator used')
        primary.header['FXCORVER'] = ('1', 'Correlator version')
        primary.header['LWATYPE'] = ('IDI-EXTENDED-ZA', 'LWA FITS file type')
        primary.header['LWAMAJV'] = (IDIVersion[0], 'LWA FITS file format major version')
        primary.header['LWAMINV'] = (IDIVersion[1], 'LWA FITS file format minor version')
        primary.header['DATE-OBS'] = (self.refTime, 'IDI file data collection date')
        ts = str(astro.get_date_from_sys())
        primary.header['DATE-MAP'] = (ts.split()[0], 'IDI file creation date')
        
        # Write the comments and history
        try:
            for comment in self._comments:
                primary.header['comment'] = comment
            del self._comments
        except AttributeError:
            pass
        primary.header['COMMENT'] = " FITS (Flexible Image Transport System) format is defined in 'Astronomy and Astrophysics', volume 376, page 359; bibcode: 2001A&A...376..359H"
        try:
            for hist in self._history:
                primary.header['history'] = hist
            del self._history
        except AttributeError:
            pass
            
        self.FITS.append(primary)
        self.FITS.flush()



# -*- coding: utf-8 -*-

# Python3 compatiability
from __future__ import division
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for lsl.reader modules"""

import os
import unittest

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import tbw
from lsl.reader import tbn
from lsl.reader import drx
from lsl.reader import vdif
from lsl.reader import drspec
from lsl.reader import errors


__revision__ = "$Rev: 2470 $"
__version__  = "0.6"
__author__    = "Jayce Dowell"


tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')
drxFile = os.path.join(dataPath, 'tests', 'drx-test.dat')
vdifFile = os.path.join(dataPath, 'tests', 'vdif-test.dat')
drspecFile = os.path.join(dataPath, 'tests', 'drspec-test.dat')


class reader_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.reader
    modules."""
    
    ### TBW ###
    
    def test_tbw_read(self):
        """Test reading in a frame from a TBW file."""
        
        fh = open(tbwFile, 'rb')
        # First frame is really TBW and stores the correct stand ID
        frame1 = tbw.readFrame(fh)
        self.assertTrue(frame1.header.isTBW())
        self.assertEqual(frame1.parseID(), 2)
        # Second frame
        frame2 = tbw.readFrame(fh)
        self.assertTrue(frame2.header.isTBW())
        self.assertEqual(frame2.parseID(), 1)
        fh.close()
        
    def test_tbw_bits(self):
        """Test getting the data bits from a TBW file."""
        
        fh = open(tbwFile, 'rb')
        # File contains 12-bit data, two ways
        self.assertEqual(tbw.getDataBits(fh), 12)
        frame1 = tbw.readFrame(fh)
        self.assertEqual(frame1.getDataBits(), 12)
        fh.close()
        
    def test_tbw_errors(self):
        """Test reading errors."""
        
        fh = open(tbwFile, 'rb')
        # Frames 1 through 8
        for i in range(1,9):
            frame = tbw.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, tbw.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(tbwFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, tbw.readFrame, fh)
        fh.close()
        
    def test_tbw_comps(self):
        """Test the TBW frame comparison operators (>, <, etc.) for time tags."""
        
        fh = open(tbwFile, 'rb')
        # Frames 1 through 3
        frames = []
        for i in range(1,4):
            frames.append(tbw.readFrame(fh))
        fh.close()
        
        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertFalse(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertFalse(frames[0] == frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_tbw_sort(self):
        """Test sorting TBW frames by time tags."""
        
        fh = open(tbwFile, 'rb')
        # Frames 1 through 3
        frames = []
        for i in range(1,4):
            frames.append(tbw.readFrame(fh))
        fh.close()
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
            
    def test_tbw_math(self):
        """Test mathematical operations on TBW frame data via frames."""
        
        fh = open(tbwFile, 'rb')
        # Frames 1 through 3
        frames = []
        for i in range(1,4):
            frames.append(tbw.readFrame(fh))
        fh.close()
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(800):
            self.assertAlmostEqual(frameT.data.xy[i%2, i//2], 2*frames[0].data.xy[i%2, i//2], 2)
        frameT *= 2.0
        for i in range(800):
            self.assertAlmostEqual(frameT.data.xy[i%2, i//2], 4*frames[0].data.xy[i%2, i//2], 2)
        frameT = frames[0] * frames[1]
        for i in range(800):
            self.assertAlmostEqual(frameT.data.xy[i%2, i//2], frames[0].data.xy[i%2, i//2]*frames[1].data.xy[i%2, i//2], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(800):
            self.assertAlmostEqual(frameA.data.xy[i%2, i//2], 2+frames[0].data.xy[i%2, i//2], 2)
        frameA += 2.0
        for i in range(800):
            self.assertAlmostEqual(frameA.data.xy[i%2, i//2], 4+frames[0].data.xy[i%2, i//2], 2)
        frameA = frames[0] + frames[1]
        for i in range(800):
            self.assertAlmostEqual(frameA.data.xy[i%2, i//2], frames[0].data.xy[i%2, i//2]+frames[1].data.xy[i%2, i//2], 2)
            
    ### TBN ###
    
    def test_tbn_read(self):
        """Test reading in a frame from a TBN file."""
        
        fh = open(tbnFile, 'rb')
        # First frame is really TBN and stores the correct IDs
        frame1 = tbn.readFrame(fh)
        stand, pol = frame1.parseID()
        self.assertEqual(stand, 1)
        self.assertEqual(pol, 0)
        # Second frame
        frame2 = tbn.readFrame(fh)
        stand, pol = frame2.parseID()
        self.assertEqual(stand, 1)
        self.assertEqual(pol, 1)
        fh.close()
        
    def test_tbn_errors(self):
        """Test reading in all frames from a truncated TBN file."""
        
        fh = open(tbnFile, 'rb')
        # Frames 1 through 29
        for i in range(1,30):
            frame = tbn.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, tbn.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(tbnFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, tbn.readFrame, fh)
        fh.close()
        
    def test_tbn_block(self):
        """Test finding out how many stands are in a file."""
        
        fh = open(tbnFile, 'rb')
        nx, ny = tbn.getFramesPerObs(fh)
        self.assertEqual(nx, 10)
        self.assertEqual(ny, 10)
        fh.close()
        
    def test_tbn_rate(self):
        """Test finding out the sample rate of a TBN file."""
        
        fh = open(tbnFile, 'rb')
        rate = tbn.getSampleRate(fh)
        self.assertEqual(rate, 100000)
        code = tbn.getSampleRate(fh, FilterCode=True)
        self.assertEqual(code, 7)
        fh.close()
        
    def test_tbn_comps(self):
        """Test the TBN frame comparison operators (>, <, etc.) for time tags."""
        
        fh = open(tbnFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(tbn.readFrame(fh))
        fh.close()
        
        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertFalse(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertFalse(frames[0] == frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_tbn_sort(self):
        """Test sorting TBN frames by time tags."""
        
        fh = open(tbnFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(tbn.readFrame(fh))
        fh.close()
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
            
    def test_tbn_math(self):
        """Test mathematical operations on TBN frame data via frames."""
        
        fh = open(tbnFile, 'rb')
        # Frames 1 through 29
        frames = []
        for i in range(1,30):
            frames.append(tbn.readFrame(fh))
        fh.close()
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(512):
            self.assertAlmostEqual(frameT.data.iq[i], 2*frames[0].data.iq[i], 2)
        frameT *= 2.0
        for i in range(512):
            self.assertAlmostEqual(frameT.data.iq[i], 4*frames[0].data.iq[i], 2)
        frameT = frames[0] * frames[1]
        for i in range(512):
            self.assertAlmostEqual(frameT.data.iq[i], frames[0].data.iq[i]*frames[1].data.iq[i], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(512):
            self.assertAlmostEqual(frameA.data.iq[i], 2+frames[0].data.iq[i], 2)
        frameA += 2.0
        for i in range(512):
            self.assertAlmostEqual(frameA.data.iq[i], 4+frames[0].data.iq[i], 2)
        frameA = frames[0] + frames[1]
        for i in range(512):
            self.assertAlmostEqual(frameA.data.iq[i], frames[0].data.iq[i]+frames[1].data.iq[i], 2)
            
    ### TBW/TBN Mix-up ###
    
    def test_tbw_tbn_catch(self):
        """Test that tbw will not read tbn files and vice versa."""
        
        fh = open(tbnFile, 'rb')
        frame1 = tbw.readFrame(fh)
        self.assertFalse(frame1.header.isTBW())
        fh.close()
        
        fh = open(tbwFile, 'rb')
        frame1 = tbn.readFrame(fh)
        self.assertFalse(frame1.header.isTBN())
        fh.close()
        
    ### DRX ###
    
    def test_drx_read(self):
        """Test reading in a frame from a DRX file."""
        
        fh = open(drxFile, 'rb')
        # First frame is really DRX and stores the IDs
        frame1 = drx.readFrame(fh)
        beam, tune, pol = frame1.parseID()
        self.assertEqual(beam, 4)
        self.assertEqual(tune, 1)
        self.assertEqual(pol,  1)
        # Second frame
        frame2 = drx.readFrame(fh)
        beam, tune, pol = frame2.parseID()
        self.assertEqual(beam, 4)
        self.assertEqual(tune, 2)
        self.assertEqual(pol,  0)
        fh.close()
        
    def test_drx_errors(self):
        """Test reading in all frames from a truncated DRX file."""
        
        fh = open(drxFile, 'rb')
        # Frames 1 through 32
        for i in range(1,33):
            frame = drx.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, drx.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(drxFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, drx.readFrame, fh)
        fh.close()
        
    def test_drx_beam(self):
        """Test finding out how many beams are present in a DRX file."""
        
        fh = open(drxFile, 'rb')
        nBeam = drx.getBeamCount(fh)
        self.assertEqual(nBeam, 1)
        fh.close()
        
    def test_drx_block(self):
        """Test finding out how many tunings/pols. per beam are in a DRX file."""
        
        fh = open(drxFile, 'rb')
        b1, b2, b3, b4 = drx.getFramesPerObs(fh)
        self.assertEqual(b1, 0)
        self.assertEqual(b2, 0)
        self.assertEqual(b3, 0)
        self.assertEqual(b4, 4)
        fh.close()
        
    def test_drx_rate(self):
        """Test finding out the DRX sample rate."""
        
        fh = open(drxFile, 'rb')
        cFrame = drx.readFrame(fh)
        fh.seek(0)
        
        # Sample rate
        self.assertEqual(cFrame.getSampleRate(), drx.getSampleRate(fh))
        
        # Filter code
        self.assertEqual(cFrame.getFilterCode(), drx.getSampleRate(fh, FilterCode=True))
        fh.close()
        
    def test_drx_comps(self):
        """Test the DRX frame comparison operators (>, <, etc.) for time tags."""
        
        fh = open(drxFile, 'rb')
        # Frames 1 through 10
        frames = []
        for i in range(1,11):
            frames.append(drx.readFrame(fh))
        fh.close()
        
        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertFalse(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertFalse(frames[0] == frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_drx_sort(self):
        """Test sorting DRX frames by time tags."""
        
        fh = open(drxFile, 'rb')
        # Frames 1 through 10
        frames = []
        for i in range(1,11):
            frames.append(drx.readFrame(fh))
        fh.close()
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
            
    def test_drx_math(self):
        """Test mathematical operations on DRX frame data via frames."""
        
        fh = open(drxFile, 'rb')
        # Frames 1 through 10
        frames = []
        for i in range(1,11):
            frames.append(drx.readFrame(fh))
        fh.close()
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(4096):
            self.assertAlmostEqual(frameT.data.iq[i], 2*frames[0].data.iq[i], 2)
        frameT *= 2.0
        for i in range(4096):
            self.assertAlmostEqual(frameT.data.iq[i], 4*frames[0].data.iq[i], 2)
        frameT = frames[0] * frames[1]
        for i in range(4096):
            self.assertAlmostEqual(frameT.data.iq[i], frames[0].data.iq[i]*frames[1].data.iq[i], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(4096):
            self.assertAlmostEqual(frameA.data.iq[i], 2+frames[0].data.iq[i], 2)
        frameA += 2.0
        for i in range(4096):
            self.assertAlmostEqual(frameA.data.iq[i], 4+frames[0].data.iq[i], 2)
        frameA = frames[0] + frames[1]
        for i in range(4096):
            self.assertAlmostEqual(frameA.data.iq[i], frames[0].data.iq[i]+frames[1].data.iq[i], 2)
            
    ### DR Spectrometer ###
    
    def test_drspec_read(self):
        """Test reading in a frame from a DR spectrometer file."""
        
        fh = open(drspecFile, 'rb')
        # First frame is really DR spectrometer and stores the IDs
        frame1 = drspec.readFrame(fh)
        beam = frame1.parseID()
        self.assertEqual(beam, 1)
        
        # Second frame
        frame2 = drspec.readFrame(fh)
        beam = frame2.parseID()
        self.assertEqual(beam, 1)
        fh.close()
        
    def test_drspec_errors(self):
        """Test reading in all frames from a truncated DR spectrometer file."""
        
        fh = open(drspecFile, 'rb')
        # Frames 1 through 8
        for i in range(1,8):
            frame = drspec.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, drspec.readFrame, fh)
        fh.close()
        
        # If we offset in the file by 1 byte, we should be a 
        # sync error (errors.syncError).
        fh = open(drspecFile, 'rb')
        fh.seek(1)
        self.assertRaises(errors.syncError, drspec.readFrame, fh)
        fh.close()
        
    def test_drspec_metadata(self):
        """Test finding out the DR spectrometer metadata."""
        
        fh = open(drspecFile, 'rb')
        cFrame = drspec.readFrame(fh)
        fh.seek(0)
        
        # Beam
        self.assertEqual(cFrame.parseID(), 1)
        
        # Sample rate
        self.assertAlmostEqual(cFrame.getSampleRate(), 19.6e6, 1)
        self.assertAlmostEqual(cFrame.getSampleRate(), drspec.getSampleRate(fh), 1)
        
        # Filter code
        self.assertEqual(cFrame.getFilterCode(), 7)
        self.assertEqual(cFrame.getFilterCode(), drspec.getSampleRate(fh, FilterCode=True))
        
        # FFT windows per integration
        self.assertEqual(cFrame.getFFTsPerIntegration(), 6144)
        self.assertEqual(cFrame.getFFTsPerIntegration(), drspec.getFFTsPerIntegration(fh))
        
        # Transform size
        self.assertEqual(cFrame.getTransformSize(), 1024)
        self.assertEqual(cFrame.getTransformSize(), drspec.getTransformSize(fh))
        
        # Integration time
        self.assertAlmostEqual(cFrame.getIntegrationTime(), 0.32099265, 8)
        self.assertAlmostEqual(cFrame.getIntegrationTime(), drspec.getIntegrationTime(fh), 8)
        
        fh.close()
        
    def test_drspec_comps(self):
        """Test the DR spectrometer frame comparison operators (>, <, etc.) for time tags."""

        fh = open(drspecFile, 'rb')
        # Frames 1 through 7
        frames = []
        for i in range(1,8):
            frames.append(drspec.readFrame(fh))
        fh.close()

        self.assertTrue(0 < frames[0])
        self.assertFalse(0 > frames[0])
        self.assertTrue(frames[-1] >= frames[0])
        self.assertFalse(frames[-1] <= frames[0])
        self.assertTrue(frames[0] == frames[0])
        self.assertFalse(frames[0] == frames[-1])
        self.assertFalse(frames[0] != frames[0])
        
    def test_drspec_sort(self):
        """Test sorting DR spectrometer frames by time tags."""
        
        fh = open(drspecFile, 'rb')
        # Frames 1 through 7
        frames = []
        for i in range(1,8):
            frames.append(drspec.readFrame(fh))
        
        frames.sort()
        frames = frames[::-1]
        
        for i in xrange(1,len(frames)):
            self.assertTrue( frames[i-1] >= frames[i] )
        fh.close()
        
    def test_drspec_math(self):
        """Test mathematical operations on DR spectrometer frame data via frames."""
        
        fh = open(drspecFile, 'rb')
        # Frames 1 through 7
        frames = []
        for i in range(1,8):
            frames.append(drspec.readFrame(fh))
        fh.close()
        
        nPts = frames[0].data.XX0.size
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in xrange(nPts):
            self.assertAlmostEqual(frameT.data.XX0[i], 2*frames[0].data.XX0[i], 2)
        frameT *= 2.0
        for i in xrange(nPts):
            self.assertAlmostEqual(frameT.data.XX1[i], 4*frames[0].data.XX1[i], 2)
        frameT = frames[0] * frames[1]
        for i in xrange(nPts):
            self.assertAlmostEqual(frameT.data.YY0[i], frames[0].data.YY0[i]*frames[1].data.YY0[i], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in xrange(nPts):
            self.assertAlmostEqual(frameA.data.XX0[i], 2+frames[0].data.XX0[i], 2)
        frameA += 2.0
        for i in xrange(nPts):
            self.assertAlmostEqual(frameA.data.XX1[i], 4+frames[0].data.XX1[i], 2)
        frameA = frames[0] + frames[1]
        for i in xrange(nPts):
            self.assertAlmostEqual(frameA.data.YY0[i], frames[0].data.YY0[i]+frames[1].data.YY0[i], 2)
            
    ### VDIF ###
    
    def test_vdif_read(self):
        """Test reading in a frame from a VDIF file."""
        
        fh = open(vdifFile, 'rb')
        # First frame
        frame1 = vdif.readFrame(fh)
        
        # Validate header
        station, thread = frame1.parseID()
        self.assertEqual(station, 19284)
        self.assertEqual(thread, 0)
        self.assertEqual(frame1.header.isLegacy, 0)
        self.assertEqual(frame1.header.isInvalid, 0)
        self.assertEqual(frame1.header.refEpoch, 30)
        self.assertEqual(frame1.header.secondsFromEpoch, 9106862)
        self.assertEqual(frame1.header.frameInSecond, 59866)
        self.assertEqual(frame1.header.frameLength, 8224//8)
        self.assertEqual(frame1.header.nChan, 4)
        self.assertEqual(frame1.header.bitsPerSample, 2)
        self.assertEqual(frame1.header.isComplex, 0)
        
        # Validate (some) data
        for k,d in enumerate((1.0, -1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0 ,-1.0)):
            i = k % frame1.header.nChan
            j = k // frame1.header.nChan
            self.assertAlmostEqual(frame1.data.data[i,j], d, 5)
            
        # Second frame
        frame2 = vdif.readFrame(fh)
        
        # Validate header
        station, thread = frame2.parseID()
        self.assertEqual(station, 19284)
        self.assertEqual(thread, 0)
        self.assertEqual(frame2.header.isLegacy, 0)
        self.assertEqual(frame2.header.isInvalid, 0)
        self.assertEqual(frame2.header.refEpoch, 30)
        self.assertEqual(frame2.header.secondsFromEpoch, 9106862)
        self.assertEqual(frame2.header.frameInSecond, 59867)
        self.assertEqual(frame2.header.frameLength, 8224//8)
        self.assertEqual(frame2.header.nChan, 4)
        self.assertEqual(frame2.header.bitsPerSample, 2)
        self.assertEqual(frame2.header.isComplex, 0)
        
        # Validate (some) data
        for k,d in enumerate((-1.0, 1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0 ,1.0)):
            i = k % frame2.header.nChan
            j = k // frame2.header.nChan
            self.assertAlmostEqual(frame2.data.data[i,j], d, 5)
            
        fh.close()
        
    def test_vdif_errors(self):
        """Test reading in all frames from a truncated VDIF file."""
        
        fh = open(vdifFile, 'rb')
        # Frames 1 through 10
        for i in range(1,11):
            frame = vdif.readFrame(fh)
            
        # Last frame should be an error (errors.eofError)
        self.assertRaises(errors.eofError, vdif.readFrame, fh)
        fh.close()
        
    def test_vdif_threads(self):
        """Test finding out how many threads file."""
        
        fh = open(vdifFile, 'rb')
        nt = vdif.getThreadCount(fh)
        self.assertEqual(nt, 1)
        fh.close()
        
    def test_vdif_math(self):
        """Test mathematical operations on VDIF frame data via frames."""
        
        fh = open(vdifFile, 'rb')
        # Frames 1 through 10
        frames = []
        for i in range(1,11):
            frames.append(vdif.readFrame(fh))
        fh.close()
        
        nChan, nSamples = frames[0].data.data.shape
        
        # Multiplication
        frameT = frames[0] * 2.0
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameT.data.data[i,j], 2*frames[0].data.data[i,j], 2)
        frameT *= 2.0
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameT.data.data[i,j], 4*frames[0].data.data[i,j], 2)
        frameT = frames[0] * frames[1]
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameT.data.data[i,j], frames[0].data.data[i,j]*frames[1].data.data[i,j], 2)
            
        # Addition
        frameA = frames[0] + 2.0
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameA.data.data[i,j], 2+frames[0].data.data[i,j], 2)
        frameA += 2.0
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameA.data.data[i,j], 4+frames[0].data.data[i,j], 2)
        frameA = frames[0] + frames[1]
        for i in range(nChan):
            for j in range(nSamples):
                self.assertAlmostEqual(frameA.data.data[i,j], frames[0].data.data[i,j]+frames[1].data.data[i,j], 2)


class reader_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.reader units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(reader_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""
Python module to read in VDIF data.  This module defines the following 
classes for storing the VIDF data found in a file:

Frame
  object that contains all data associated with a particular DRX frame.  
  The primary constituents of each frame are:
    * FrameHeader - the VDIF frame header object and
    * FrameData   - the VDIF frame data object.
Combined, these two objects contain all of the information found in the 
original VDIF frame.

The functions defined in this module fall into two class:
  1. convert a frame in a file to a Frame object and
  2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.  The readBlock
function reads in a (user-defined) number of VDIF frames and returns a 
ObservingBlock object.

For describing the format of data in the file, two function are provided:

getThreadCount
  read in the first few frames of an open file handle and return how many 
  threads are present in the file.
"""

import copy
from datetime import datetime

from lsl import astro
from lsl.common.mcs import datetime2mjdmpm
from lsl.reader._gofast import readVDIF
from lsl.reader._gofast import syncError as gsyncError
from lsl.reader._gofast import eofError as geofError
from lsl.reader.errors import syncError, eofError


__version__ = '0.4'
__revision__ = '$Rev: 2581 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'readGUPPIHeader', 'readFrame', 
           'getFrameSize', 'getThreadCount', 'getFramesPerSecond', 'getSampleRate', 
           '__version__', '__revision__', '__all__']



def _crcc(data, length=48, mask=0o40003, cycle=16):
    """
    Compute a CRC checksum for the VLBA BCD time code stored in a Mark 5B 
    header.
    
    From:  mk5blib.c
    """
    
    state = 0
    for i in xrange(length):
        q = state & 1
        if ((data >> i) & 1) ^ q == 0:
            state &= -2
        else:
            state ^= mask
            state |= i
        state = (state >> 1) | (state & 1) << (cycle-1)
    return state


class FrameHeader(object):
    """
    Class that stores the information found in the header of a VDIF 
    frame.  Most fields in the VDIF version 1.1.1 header are stored.
    """
    
    def __init__(self, isInvalid=0, isLegacy=0, secondsFromEpoch=0, refEpoch=0, frameInSecond=0, version=1, nChan=0, frameLength=0, isComplex='C', bitsPerSample=0, threadID=0, stationID=0, extendedData1=None, extendedData2=None, extendedData3=None, extendedData4=None, sampleRate=0.0, centralFreq=0.0):
        self.isInvalid = isInvalid
        self.isLegacy = isLegacy
        self.secondsFromEpoch = secondsFromEpoch
        
        self.refEpoch = refEpoch
        self.frameInSecond = frameInSecond
        
        self.version = version
        self.nChan = nChan
        self.frameLength = frameLength
        
        self.isComplex = isComplex
        self.bitsPerSample = bitsPerSample
        self.threadID = threadID
        self.stationID = stationID
        
        self.extendedData1 = extendedData1
        self.extendedData2 = extendedData2
        self.extendedData3 = extendedData3
        self.extendedData4 = extendedData4
        
        self.sampleRate = sampleRate
        self.centralFreq = centralFreq
        
    def getTime(self):
        """
        Function to convert the time tag to seconds since the UNIX epoch.
        """
        
        # Get the reference epoch in the strange way that it is stored in VDIF 
        # and convert it to a MJD
        epochDT = datetime(2000+self.refEpoch/2, (self.refEpoch % 2)*6+1, 1, 0, 0, 0, 0)
        epochMJD, epochMPM = datetime2mjdmpm(epochDT)
        epochMJD = epochMJD + epochMPM/1000.0/86400.0
        
        # Get the frame MJD by adding the secondsFromEpoch value to the epoch
        frameMJD_i = epochMJD + self.secondsFromEpoch / 86400
        frameMJD_f = (self.secondsFromEpoch % 86400) / 86400.0
        
        if self.sampleRate == 0.0:
            # Try to get the sub-second time by parsing the extended user data
            try:
                ## Is there a sample rate to grab?
                eud = self.parseExtendedUserData()
                sampleRate = eud['sampleRate']
                sampleRate *= 1e6 if eud['sampleRateUnits'] == 'MHz' else 1e3
                
                ## How many samples are in each frame?
                dataSize = self.frameLength*8 - 32 + 16*self.isLegacy		# 8-byte chunks -> bytes - full header + legacy offset
                samplesPerWord = 32 / self.bitsPerSample					# dimensionless
                nSamples = dataSize / 4 * samplesPerWord					# bytes -> words -> samples
            
                ## What is the frame rate?
                frameRate = sampleRate / nSamples
                
                frameMJD_f += 1.0*self.frameInSecond/frameRate/86400.0
                
            except KeyError:
                pass
                
        else:
            # Use what we already have been told
            ## How many samples are in each frame?
            dataSize = self.frameLength*8 - 32 + 16*self.isLegacy		# 8-byte chunks -> bytes - full header + legacy offset
            samplesPerWord = 32 / self.bitsPerSample				# dimensionless
            nSamples = dataSize / 4 * samplesPerWord				# bytes -> words -> samples
        
            ## What is the frame rate?
            frameRate = self.sampleRate / nSamples
            
            frameMJD_f += 1.0*self.frameInSecond/frameRate/86400.0

        # Convert from MJD to UNIX time
        if frameMJD_f > 1:
            frameMJD_i += 1
            frameMJD_f -= 1
        seconds = int(astro.utcjd_to_unix(frameMJD_i + astro.MJD_OFFSET))
        seconds += frameMJD_f * 86400.0
        
        return seconds
        
    def parseID(self):
        """
        Return a two-element tuple of the station ID and thread ID.
        
        .. note::
            The station ID is always returned as numeric.
        """
        return (self.stationID, self.threadID)
        
    def parseExtendedUserData(self):
        """
        Parse the extended user data if it was included with the reader.  
        The data is returned as a dictionary.
        """
        
        # Setup the output dictionary
        fields = {}
        
        # Is there anything to look at?
        if self.extendedData1 is None or self.extendedData2 is None or self.extendedData3 is None or self.extendedData4 is None:
            return fields
            
        # Extract the version
        edv = int((self.extendedData1 >> 24) & 0xFF)
        
        # Parse accordingly
        if edv == 0x00:
            ## Empty
            pass
            
        elif edv == 0x01:
            ## NICT
            fields['sampleRate'] = int(self.extendedData1 & (2**23-1))
            fields['sampleRateUnits'] = 'MHz' if int((self.extendedData1>>23) & 1) else 'kHz'
            fields['syncPattern'] = self.extendedData2
            fields['stationName'] = (self.extendedData4 << 32) | self.extendedData3
            
        elif edv == 0x02:
            ## ALMA
            fields['syncWord'] = int(self.extendedData1 & 0xFFFF)
            fields['picStatusWord'] = self.extendedData2
            fields['packetSerialNumber'] = (self.extendedData4 << 32) | self.extendedData3
            
        elif edv == 0x03:
            ## NRAO
            fields['sampleRate'] = int(self.extendedData1 & (2**23-1))
            fields['sampleRateUnits'] = 'MHz' if int((self.extendedData1 >> 23) & 1) else 'kHz'
            fields['syncPattern'] = self.extendedData2
            fields['tuningWord'] = self.extendedData3
            fields['dbeUnit'] = int((self.extendedData4 >> 24) & 0xF)
            fields['ifInputNumber'] = int((self.extendedData4 >> 20) & 0xF)
            fields['subBand'] = int((self.extendedData4 >> 17) & 0x7)
            fields['electronicSideBand'] = 'USB' if (self.extendedData4 >> 16) & 1 else 'LSB'
            mj = int((self.extendedData4 >> 12) & 0xF)
            mn = int((self.extendedData4 >>  8) & 0xF)
            pr = int(self.extendedData4 & 0xFF)
            fields['version'] = '%i.%i-%02f' % (mj, mn, pr)
            
        elif edv == 0xAB:
            ## Haystack (which is really an embedded Mark 5B header)
            fields['syncWord'] = self.extendedData1
            fields['yearsFrom2000'] = int((self.extendedData2 >> 28) & 0xF)
            fields['userSpecifiedData'] = int((self.extendedData2 >> 16) & 0xFFF)
            fields['dataFromInternalTVG'] = (self.extendedData2 >> 15) & 1
            fields['frameInSecond'] = int(self.extendedData2 & (2**14-1))
            j0 = int((self.extendedData3 >> 28) & 0xF)
            j1 = int((self.extendedData3 >> 24) & 0xF)
            j2 = int((self.extendedData3 >> 20) & 0xF)
            s0 = int((self.extendedData3 >> 16) & 0xF)
            s1 = int((self.extendedData3 >> 12) & 0xF)
            s2 = int((self.extendedData3 >>  8) & 0xF)
            s3 = int((self.extendedData3 >>  4) & 0xF)
            s4 = int((self.extendedData3 >>  0) & 0xF)
            f0 = int((self.extendedData4 >> 28) & 0xF)
            f1 = int((self.extendedData4 >> 24) & 0xF)
            f2 = int((self.extendedData4 >> 20) & 0xF)
            f3 = int((self.extendedData4 >> 16) & 0xF)
            f4 = int((self.extendedData4 >>  8) & 0xF)
            crcf = int(self.extendedData4 & 0xFFFF)
            crcf = (crcf & 0xF) << 8 | ((crcf >> 8) & 0xF)
            crcc = _crcc((self.extendedData3 << 16) | ((self.extendedData3 >> 16) & 0xFF), 48)
            fields['vlbaTimeCode'] = j0*1e7+j1*1e6+j2*1e5 + s0*1e4+s1*1e3+s2*1e2+s3*1e1+s4*1e0 + f0/1e1+f1/1e2+f2/1e3+f3/1e4+f4/1e5
            fields['vlbaTimeCodeValue'] = True if crcc == crcf else False
            
        else:
            raise RuntimeError("Unknown extended user data version: %i" % edv)
            
        return fields
        
    def getSampleRate(self):
        """
        Return the sample rate of the data in samples/second.
        """
        
        return self.sampleRate*1.0
        
    def getCentralFreq(self):
        """
        Function to get the central frequency of the VDIF data in Hz.
        """
        
        return self.centralFreq*1.0


class FrameData(object):
    """
    Class that stores the information found in the data section of a VDIF
    frame.
    
    .. note::
        Unlike the other readers in the :mod:`lsl.reader` module the
        data are stored as numpy.float32 values.
    """
    
    def __init__(self, data=None):
        self.data = data


class Frame(object):
    """
    Class that stores the information contained within a single VDIF 
    frame.  It's properties are FrameHeader and FrameData objects.
    """

    def __init__(self, header=None, data=None):
        if header is None:
            self.header = FrameHeader()
        else:
            self.header = header
            
        if data is None:
            self.data = FrameData()
        else:
            self.data = data
            
        self.valid = True

    def parseID(self):
        """
        Convenience wrapper for the Frame.FrameHeader.parseID 
        function.
        """
        
        return self.header.parseID()
        
    def parseExtendedUserData(self):
        """
        Convenience wrapper for the Frame.FrameHeader.parseExtendedUserData
        function.
        """
        
        return self.header.parseExtendedUserData()
        
    def getTime(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getTime function.
        """
        
        return self.header.getTime()
        
    def getSampleRate(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getSampleRate function.
        """
        
        return self.header.getSampleRate()
        
    def getCentralFreq(self):
        """
        Convenience wrapper for the Frame.FrameHeader.getCentralFreq function.
        """
        
        return self.header.getCentralFreq()
        
    def __add__(self, y):
        """
        Add the data sections of two frames together or add a number 
        to every element in the data section.
        """
    
        newFrame = copy.deepcopy(self)
        newFrame += y	
        return newFrame
            
    def __iadd__(self, y):
        """
        In-place add the data sections of two frames together or add 
        a number to every element in the data section.
        """
        
        try:
            self.data.data += y.data.data
        except AttributeError:
            self.data.data += y
        return self
        
    def __mul__(self, y):
        """
        Multiple the data sections of two frames together or multiply 
        a number to every element in the data section.
        """
        
        newFrame = copy.deepcopy(self)
        newFrame *= y
        return newFrame
        
    def __imul__(self, y):
        """
        In-place multiple the data sections of two frames together or 
        multiply a number to every element in the data section.
        """
        
        try:
            self.data.data *= y.data.data
        except AttributeError:
            self.data.data *= y
        return self
        
    #def __eq__(self, y):
    #	"""
    #	Check if the time tags of two frames are equal or if the time
    #	tag is equal to a particular value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX == tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __ne__(self, y):
    #	"""
    #	Check if the time tags of two frames are not equal or if the time
    #	tag is not equal to a particular value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX != tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __gt__(self, y):
    #	"""
    #	Check if the time tag of the first frame is greater than that of a
    #	second frame or if the time tag is greater than a particular value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX > tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __ge__(self, y):
    #	"""
    #	Check if the time tag of the first frame is greater than or equal to 
    #	that of a second frame or if the time tag is greater than a particular 
    #	value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX >= tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __lt__(self, y):
    #	"""
    #	Check if the time tag of the first frame is less than that of a
    #	second frame or if the time tag is greater than a particular value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX < tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __le__(self, y):
    #	"""
    #	Check if the time tag of the first frame is less than or equal to 
    #	that of a second frame or if the time tag is greater than a particular 
    #	value.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	try:
    #		tY = y.data.timeTag
    #	except AttributeError:
    #		tY = y
    #	
    #	if tX <= tY:
    #		return True
    #	else:
    #		return False
    #		
    #def __cmp__(self, y):
    #	"""
    #	Compare two frames based on the time tags.  This is helpful for 
    #	sorting things.
    #	"""
    #	
    #	tX = self.data.timeTag
    #	tY = y.data.timeTag
    #	if tY > tX:
    #		return -1
    #	elif tX > tY:
    #		return 1
    #	else:
    #		return 0


def readGUPPIHeader(filehandle):
    """
    Read in a GUPPI header at the start of a VDIF file from the VLA.  The 
    contents of the header are returned as a dictionary.
    """
    
    # Read in the GUPPI header
    header = {}
    while True:
        line = filehandle.read(80)
        if line[:3] == 'END':
            break
        elif line[:8] == 'CONTINUE':
            junk, value2 = line.split(None, 1)
            value = "%s%s" % (value[:-1], value2[:-1])
        else:
            name, value = line.split('=', 1)
            name = name.strip()
        try:
            value = int(value, 10)
        except:
            try:
                value = float(value)
            except:
                value = value.strip().replace("'", '')
        header[name.strip()] = value
    header['OBSBW'] *= 1e6
    header['OBSFREQ'] *= 1e6
    
    return header


def readFrame(filehandle, sampleRate=0.0, centralFreq=0.0, Verbose=False):
    """
    Function to read in a single VDIF frame (header+data) and store the 
    contents as a Frame object.  This function wraps the _readerHeader and 
    _readData functions.
    """
    
    # New _vdif method
    try:
        newFrame = readVDIF(filehandle, Frame(), centralFreq=centralFreq, sampleRate=sampleRate)
    except gsyncError:
        mark = filehandle.tell()
        raise syncError(location=mark)
    except geofError:
        raise eofError
        
    return newFrame


def getFrameSize(filehandle, nFrames=None):
    """
    Find out what the frame size is in bytes from a single observation.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()

    # Read in one frame
    newFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    return newFrame.header.frameLength*8


def getThreadCount(filehandle):
    """
    Find out how many thrads are present by examining the first 1024
    records.  Return the number of threads found.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Get the frame size
    FrameSize = getFrameSize(filehandle)
    
    # Build up the list-of-lists that store ID codes and loop through 1024
    # frames.  In each case, parse pull the thread ID and append the thread 
    # ID to the relevant thread array if it is not already there.
    threads = []
    i = 0
    while i < 1024:
        try:
            cFrame = readFrame(filehandle)
        except syncError:
            filehandle.seek(FrameSize, 1)
            continue
        except eofError:
            break
            
        cID = cFrame.header.threadID
        if cID not in threads:
            threads.append(cID)
        i += 1
        
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Return the number of threads found
    return len(threads)


def getFramesPerSecond(filehandle):
    """
    Find out the number of frames per second in a file by watching how the 
    headers change.  Returns the number of frames in a second.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Get the frame size
    FrameSize = getFrameSize(filehandle)
    
    # Get the number of threads
    nThreads = getThreadCount(filehandle)
    
    # Get the current second counts for all threads
    ref = {}
    i = 0
    while i < nThreads:
        try:
            cFrame = readFrame(filehandle)
        except syncError:
            filehandle.seek(FrameSize, 1)
            continue
        except eofError:
            break
            
        cID = cFrame.header.threadID
        cSC = cFrame.header.secondsFromEpoch
        ref[cID] = cSC
        i += 1
        
    # Read frames until we see a change in the second counter
    cur = {}
    fnd = []
    while True:
        ## Get a frame
        try:
            cFrame = readFrame(filehandle)
        except syncError:
            filehandle.seek(FrameSize, 1)
            continue
        except eofError:
            break
            
        ## Pull out the relevant metadata
        cID = cFrame.header.threadID
        cSC = cFrame.header.secondsFromEpoch
        cFC = cFrame.header.frameInSecond
        
        ## Figure out what to do with it
        if cSC == ref[cID]:
            ### Same second as the reference, save the frame number
            cur[cID] = cFC
        else:
            ### Different second than the reference, we've found something
            ref[cID] = cSC
            if cID not in fnd:
                fnd.append( cID )
                
        if len(fnd) == nThreads:
            break
            
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Pull out the mode
    mode = {}
    for key,value in cur.iteritems():
        try:
            mode[value] += 1
        except KeyError:
            mode[value] = 1
    best, bestValue = 0, 0
    for key,value in mode.iteritems():
        if value > bestValue:
            best = key
            bestValue = value
            
    # Correct for a zero-based counter and return
    best += 1
    return best


def getSampleRate(filehandle):
    """
    Find and return the sample rate in Hz by looking at how many frames 
    there are per second and how many samples there are in a frame.
    """
    
    # Save the current position in the file so we can return to that point
    fhStart = filehandle.tell()
    
    # Get the number of frames per second
    nFramesSecond = getFramesPerSecond(filehandle)
    
    # Read in a frame
    cFrame = readFrame(filehandle)
    
    # Return to the place in the file where we started
    filehandle.seek(fhStart)
    
    # Get the sample rate
    sampleRate = cFrame.data.data.shape[-1] * nFramesSecond
    return float(sampleRate)


# -*- coding: utf-8 -*-

"""
Module to make a blinking ASCII busy indicator.
"""

import sys
import time
import threading

__version__ = '0.1'
__revision__ = '$Rev: 2384 $'
__all__ = ['BusyIndicator', '__version__', '__revision__', '__all__']


class BusyIndicator(object):
    """
    Object to make a ASCII busy indicator for use with various long-
    run tasks that don't have a way of calculating how long they will
    take.

    Example Usage:
        >>> from busy import BusyIndicator
        >>> bi = BusyIndicator()
        >>> bi.start()
        >>> longRunningTask()
        >>> bi.stop()
    """
    
    def __init__(self, message='Busy', interval=0.5, dots=3):
        """
        Initialize the BusyIndicator class with various parameters:
        * message: message to display
        * interval: interval in seconds between message displays
        * dots: number of dots to cycle through at the end of the
            message
        """
            
        self.message = message
        self.interval = interval
        self.dots = dots
        
        self.thread = None
        self.alive = threading.Event()
        self._i = 0
        
    def start(self):
        """
        Start the indicator running.
        """
        
        if self.thread is not None:
            self.stop()
            
        self.thread = threading.Thread(target=self._run, name='indicator')
        self.thread.setDaemon(1)
        self.alive.set()
        self.thread.start()
        
    def stop(self):
        """
        Stop the indicator and display a 'Done' message.
        
        .. note::
            This can take up to one BusyIndicator.interval to complete.
        """
        
        if self.thread is not None:
            sys.stdout.write('%s%sDone%s\n' % (self.message, '.'*self._i, ' '*self.dots))
            
            self.alive.clear()
            self.thread.join()
            self.thread = None
            self._i = 0
            
    def _run(self):
        """
        Internal function used by the thread to make/change the displayed text.
        """
        
        while self.alive.isSet():
            sys.stdout.write('%s%s%s\r' % (self.message, '.'*self._i, ' '*(self.dots-self._i)))
            sys.stdout.flush()
            
            self._i += 1
            self._i %= (self.dots+1)
            time.sleep(self.interval)


# -*- coding: utf-8 -*-

# Python3 compatiability
import sys
if sys.version_info > (3,):
    xrange = range
    
"""Unit test for lsl.imaging modules"""

import os
import copy
import time
import numpy
import shutil
import tempfile
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from lsl import astro
from lsl.common.paths import dataBuild as dataPath
from lsl.imaging import utils
from lsl.imaging import selfCal
from lsl.writer.fitsidi import IDI, NumericStokes
from lsl.sim.vis import srcs as simSrcs
from lsl.common.stations import lwa1, parseSSMIF
from lsl.correlator import uvUtils

run_ms_tests = False
try:
    import casacore
    from lsl.writer.measurementset import MS
    run_ms_tests = True
except ImportError:
    pass

__revision__ = "$Rev: 2650 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


uvFile = os.path.join(dataPath, 'tests', 'uv-test.fits')
idiFile = os.path.join(dataPath, 'tests', 'idi-test.fits')
idiAltFile = os.path.join(dataPath, 'tests', 'idi-test-alt.fits')
idiSSMIFFile = os.path.join(dataPath, 'tests', 'idi-test-alt.txt')


class imaging_tests(unittest.TestCase):
    """A unittest.TestCase collection of unit tests for the lsl.imaging
    modules."""
    
    testPath = None
    
    def setUp(self):
        """Turn off all numpy warnings and create the temporary file directory."""

        numpy.seterr(all='ignore')
        self.testPath = tempfile.mkdtemp(prefix='test-imaging-', suffix='.tmp')
        
    def test_CorrelatedDataIDI(self):
        """Test the utils.CorrelatedDataIDI class."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedDataIDI(idiFile)
        
        # Dates
        self.assertEqual(idi.dateObs.strftime("%Y-%m-%dT%H:%M:%S"), "2013-03-04T20:36:26")
        
        # Stand and baseline counts
        self.assertEqual(len(idi.stands), 5)
        self.assertEqual(idi.totalBaselineCount, 5*(5+1)/2)
        
        # Basic functions (just to see that they run)
        junk = idi.getAntennaArray()
        junk = idi.getObserver()
        junk = idi.getDataSet(1)
        
        # Error checking
        self.assertRaises(RuntimeError, idi.getDataSet, 2)
        
    def test_CorrelatedDataIDI_MultiIF(self):
        """Test the utils.CorrelatedDataIDI class on a file with multiple IFs."""
        
        # Frequency range
        freq = numpy.arange(0,512)*20e6/512 + 40e6
        # Site and stands
        site = lwa1
        antennas = site.getAntennas()[0:40:2]
        
        # Set baselines and data
        blList = uvUtils.getBaselines(antennas, IncludeAuto=True, Indicies=False)
        visData = numpy.random.rand(len(blList), len(freq))
        visData = visData.astype(numpy.complex64)

        data = {'freq': freq, 'site': site, 'antennas': antennas, 'bl': blList, 'vis': visData}
        
        # Filename and time
        testTime, testFile = time.time(), os.path.join('idi-test-MultiIF.fits')
        
        # Start the file
        fits = IDI(testFile, refTime=testTime, clobber=True)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setFrequency(data['freq']+30e6)
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(astro.utcjd_to_taimjd(astro.unix_to_utcjd(testTime)), 6.0, data['bl'], 
                          numpy.concatenate([data['vis'], 10*data['vis']], axis=1))
        fits.write()
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(testFile)
        self.assertEqual(idi.freq.size, 2*data['freq'].size)
        dataDict = idi.getDataSet(1, includeAuto=True)
        
        for i in xrange(len(data['bl'])):
            for j in xrange(data['freq'].size):
                self.assertAlmostEqual(dataDict['vis']['xx'][i][j], data['vis'][i,j], 6)
                self.assertAlmostEqual(dataDict['vis']['xx'][i][j+data['freq'].size], 10*data['vis'][i,j], 6)
                
    def test_CorrelatedDataIDI_Alt(self):
        """Test the utils.CorrelatedDataIDI class on a file with an unusual telescope."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedDataIDI(idiAltFile)
        
        # Dates
        self.assertEqual(idi.dateObs.strftime("%Y-%m-%dT%H:%M:%S"), "2013-03-04T20:36:26")
        
        # Stand and baseline counts
        self.assertEqual(len(idi.stands), 5)
        self.assertEqual(idi.totalBaselineCount, 5*(5+1)/2)
        self.assertEqual(idi.integrationCount, 1)
        
        # Basic functions (just to see that they run)
        junk = idi.getAntennaArray()
        junk = idi.getObserver()
        junk = idi.getDataSet(1)
        
        # Error checking
        self.assertRaises(RuntimeError, idi.getDataSet, 2)
        
    def test_CorrelatedDataIDI_AltArrayGeometry(self):
        """Test the utils.CorrelatedDataIDI class on determing array geometry."""
        
        # Open the FITS IDI files
        idi1 = utils.CorrelatedData(idiFile)
        idi2 = utils.CorrelatedData(idiAltFile)
        
        # Dates
        self.assertEqual(idi1.dateObs.strftime("%Y-%m-%dT%H:%M:%S"), idi2.dateObs.strftime("%Y-%m-%dT%H:%M:%S"))
        
        # Stand and baseline counts
        self.assertEqual(len(idi1.stands), len(idi2.stands))
        self.assertEqual(idi1.totalBaselineCount, idi2.totalBaselineCount)
        self.assertEqual(idi1.integrationCount, idi2.integrationCount)
        
        # Check stands
        for s1,s2 in zip(idi1.stands, idi2.stands):
            self.assertEqual(s1, s2)
            
        # Check stations
        station1 = parseSSMIF(idiSSMIFFile)
        station2 = idi2.station
        self.assertAlmostEqual(station1.lat, station2.lat, 3)
        self.assertAlmostEqual(station1.lon, station2.lon, 3)
        self.assertAlmostEqual(station1.elev, station2.elev, 1)
        
        # Check antennas
        ants1 = [a for a in station1.getAntennas() if a.pol == 0]
        ants2 = station2.getAntennas()
        for a1,a2 in zip(ants1, ants2):
            self.assertEqual(a1.id, a2.id)
            self.assertEqual(a1.stand.id, a2.stand.id)
            self.assertAlmostEqual(a1.stand.x, a2.stand.x, 2)
            self.assertAlmostEqual(a1.stand.y, a2.stand.y, 2)
            self.assertAlmostEqual(a1.stand.z, a2.stand.z, 2)
            
    def test_CorrelatedDataUV(self):
        """Test the utils.CorrelatedDataUV class."""
        
        # Open the UVFITS file
        uv = utils.CorrelatedDataUV(uvFile)
        
        # Dates
        self.assertEqual(uv.dateObs.strftime("%Y-%m-%dT%H:%M:%S"), "2013-03-04T20:36:26")
        
        # Stand and baseline counts
        self.assertEqual(len(uv.stands), 5)
        self.assertEqual(uv.totalBaselineCount, 5*(5+1)/2)
        self.assertEqual(uv.integrationCount, 1)
        
        # Basic functions (just to see that they run)
        junk = uv.getAntennaArray()
        junk = uv.getObserver()
        junk = uv.getDataSet(1)
        
        # Error checking
        self.assertRaises(RuntimeError, uv.getDataSet, 2)
        
    def __initData(self):
        """Private function to generate a random set of data for writing a UVFITS
        file.  The data is returned as a dictionary with keys:
         * freq - frequency array in Hz
         * site - lwa.common.stations object
         * stands - array of stand numbers
         * bl - list of baseline pairs in real stand numbers
         * vis - array of visibility data in baseline x freq format
        """

        # Frequency range
        freq = numpy.arange(0,512)*20e6/512 + 40e6
        # Site and stands
        site = lwa1
        antennas = site.getAntennas()[0:40:2]
        
        # Set baselines and data
        blList = uvUtils.getBaselines(antennas, IncludeAuto=True, Indicies=False)
        visData = numpy.random.rand(len(blList), len(freq))
        visData = visData.astype(numpy.complex64)
        
        return {'freq': freq, 'site': site, 'antennas': antennas, 'bl': blList, 'vis': visData}
        
    @unittest.skipUnless(run_ms_tests, "requires the 'casacore' module")
    def test_CorrelatedDataMS(self):
        """Test the utils.CorrelatedDataMS class."""
        
        testTime, testFile = time.time(), os.path.join(self.testPath, 'ms-test-W.ms')
        
        # Get some data
        data = self.__initData()
        
        # Start the table
        tbl = MS(testFile, refTime=testTime)
        tbl.setStokes(['xx'])
        tbl.setFrequency(data['freq'])
        tbl.setGeometry(data['site'], data['antennas'])
        tbl.addDataSet(astro.utcjd_to_taimjd(astro.unix_to_utcjd(testTime)), 6.0, data['bl'], data['vis'])
        tbl.write()
        
        # Open the measurement set
        ms = utils.CorrelatedDataMS(testFile)
        
        # Basic functions (just to see that they run)
        junk = ms.getAntennaArray()
        junk = ms.getObserver()
        junk = ms.getDataSet(1)
        
        # Error checking
        self.assertRaises(IndexError, ms.getDataSet, 2)
        
    @unittest.skipUnless(run_ms_tests, "requires the 'casacore' module")
    def test_CorrelatedDataMS_SingleIF(self):
        """Test the utils.CorrelatedDataMS class on a file with a single IF."""
        
        # Get some data
        data = self.__initData()
        
        # Filename and time
        testTime, testFile = time.time(), os.path.join(self.testPath, 'ms-test-SingleIF.ms')
        
        # Start the file
        fits = MS(testFile, refTime=testTime, clobber=True)
        fits.setStokes(['xx'])
        fits.setFrequency(data['freq'])
        fits.setGeometry(data['site'], data['antennas'])
        fits.addDataSet(astro.utcjd_to_taimjd(astro.unix_to_utcjd(testTime)), 6.0, data['bl'], data['vis'])
        fits.write()
        
        # Open the measurement set
        ms = utils.CorrelatedDataMS(testFile)
        self.assertEqual(ms.freq.size, data['freq'].size)
        ds = ms.getDataSet(1, includeAuto=True)
        
        for i in xrange(len(data['bl'])):
            for j in xrange(data['freq'].size):
                self.assertAlmostEqual(ds['vis']['xx'][i][j], data['vis'][i,j], 6)
                
    def test_sort(self):
        """Test the utils.sortDataDict function."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiFile)
        
        # Get some data to sort
        ds = idi.getDataSet(1, sort=False)
        
        # Sort
        dss = copy.deepcopy(ds)
        utils.sortDataDict(dss)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dss[prop][pol]), len(ds[prop][pol]))
                
    def test_sort_alt(self):
        """Test the utils.sortDataDict function - alternate FITS IDI file."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiAltFile)
        
        # Get some data to sort
        ds = idi.getDataSet(1, sort=False)
        
        # Sort
        dss = copy.deepcopy(ds)
        utils.sortDataDict(dss)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dss[prop][pol]), len(ds[prop][pol]))
                
    def test_sort_uvfits(self):
        """Test the utils.sortDataDict function - UVFITS file."""
        
        # Open the FITS IDI file
        uv = utils.CorrelatedData(uvFile)
        
        # Get some data to sort
        ds = uv.getDataSet(1, sort=False)
        
        # Sort
        dss = copy.deepcopy(ds)
        utils.sortDataDict(dss)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dss[prop][pol]), len(ds[prop][pol]))
                
    def test_prune(self):
        """Test the utils.pruneBaselineRange function."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiFile)
        
        # Get some data to sort
        ds = idi.getDataSet(1)
        
        # Prune
        dsp1 = utils.pruneBaselineRange(ds, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertTrue(len(dsp1[prop][pol]) < len(ds[prop][pol]))
                
        # Auto-prune
        dsp2 = idi.getDataSet(1, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp1[prop][pol]), len(dsp2[prop][pol]))

        # Auto-prune that should result in no baselines
        dsp3 = idi.getDataSet(1, uvMin=100)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp3[prop][pol]), 0)
                
    def test_prune_alt(self):
        """Test the utils.pruneBaselineRange function - alternate FITS IDI file."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiAltFile)
        
        # Get some data to sort
        ds = idi.getDataSet(1)
        
        # Prune
        dsp1 = utils.pruneBaselineRange(ds, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertTrue(len(dsp1[prop][pol]) < len(ds[prop][pol]))
                
        # Auto-prune
        dsp2 = idi.getDataSet(1, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp1[prop][pol]), len(dsp2[prop][pol]))

        # Auto-prune that should result in no baselines
        dsp3 = idi.getDataSet(1, uvMin=100)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp3[prop][pol]), 0)
                
    def test_prune_uvfits(self):
        """Test the utils.pruneBaselineRange function - UVFITS file."""
        
        # Open the FITS IDI file
        uv = utils.CorrelatedData(uvFile)
        
        # Get some data to sort
        ds = uv.getDataSet(1)
        
        # Prune
        dsp1 = utils.pruneBaselineRange(ds, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertTrue(len(dsp1[prop][pol]) < len(ds[prop][pol]))
                
        # Auto-prune
        dsp2 = uv.getDataSet(1, uvMin=10)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp1[prop][pol]), len(dsp2[prop][pol]))

        # Auto-prune that should result in no baselines
        dsp3 = uv.getDataSet(1, uvMin=100)
        for prop in ['bls', 'uvw', 'vis', 'wgt', 'msk', 'jd']:
            for pol in ds['bls'].keys():
                self.assertEqual(len(dsp3[prop][pol]), 0)
                
    def test_rephase(self):
        """Test the utils.rephaseData function."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiFile)
        
        # Get the AntennaArray instance
        aa = idi.getAntennaArray()
        
        # Get some data to sort
        ds = idi.getDataSet(1)
        
        # Rephase #1
        rs1 = utils.rephaseData(aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['Sun'])
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs1['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs1['bls']['xx'][i][1])
            
        # Rephase #2
        rs2 = utils.rephaseData(aa, rs1, currentPhaseCenter=simSrcs['Sun'], newPhaseCenter='z')
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs2['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs2['bls']['xx'][i][1])
            
            for j in xrange(len(ds['vis']['xx'][i])):
                self.assertAlmostEqual(ds['vis']['xx'][i][j], rs2['vis']['xx'][i][j], 2)
                
        # Bad rephase
        self.assertRaises(RuntimeError, utils.rephaseData, aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['vir'])
        
    def test_rephase_alt(self):
        """Test the utils.rephaseData function - alternate FITS IDI file."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiAltFile)
        
        # Get the AntennaArray instance
        aa = idi.getAntennaArray()
        
        # Get some data to sort
        ds = idi.getDataSet(1)
        
        # Rephase #1
        rs1 = utils.rephaseData(aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['Sun'])
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs1['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs1['bls']['xx'][i][1])
            
        # Rephase #2
        rs2 = utils.rephaseData(aa, rs1, currentPhaseCenter=simSrcs['Sun'], newPhaseCenter='z')
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs2['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs2['bls']['xx'][i][1])
            
            for j in xrange(len(ds['vis']['xx'][i])):
                self.assertAlmostEqual(ds['vis']['xx'][i][j], rs2['vis']['xx'][i][j], 2)
                
        # Bad rephase
        self.assertRaises(RuntimeError, utils.rephaseData, aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['vir'])
        
    def test_rephase_uvfits(self):
        """Test the utils.rephaseData function - UVFITS file."""
        
        # Open the UVFITS file
        uv = utils.CorrelatedData(uvFile)
        
        # Get the AntennaArray instance
        aa = uv.getAntennaArray()
        
        # Get some data to sort
        ds = uv.getDataSet(1)
        
        # Rephase #1
        rs1 = utils.rephaseData(aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['Sun'])
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs1['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs1['bls']['xx'][i][1])
            
        # Rephase #2
        rs2 = utils.rephaseData(aa, rs1, currentPhaseCenter=simSrcs['Sun'], newPhaseCenter='z')
        for i in xrange(len(ds['bls']['xx'])):
            self.assertEqual(ds['bls']['xx'][i][0], rs2['bls']['xx'][i][0])
            self.assertEqual(ds['bls']['xx'][i][1], rs2['bls']['xx'][i][1])
            
            for j in xrange(len(ds['vis']['xx'][i])):
                self.assertAlmostEqual(ds['vis']['xx'][i][j], rs2['vis']['xx'][i][j], 2)
                
        # Bad rephase
        self.assertRaises(RuntimeError, utils.rephaseData, aa, ds, currentPhaseCenter='z', newPhaseCenter=simSrcs['vir'])
        
    def test_gridding(self):
        """Test building a image from a visibility data set."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiFile)
        
        # Build the image
        ds = idi.getDataSet(1)
        junk = utils.buildGriddedImage(ds, verbose=False)

        # Error checking
        self.assertRaises(RuntimeError, utils.buildGriddedImage, ds, pol='xy')
        
    def test_gridding_alt(self):
        """Test building a image from a visibility data set - alternate FITS IDI file."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiAltFile)
        
        # Build the image
        ds = idi.getDataSet(1)
        junk = utils.buildGriddedImage(ds, verbose=False)

        # Error checking
        self.assertRaises(RuntimeError, utils.buildGriddedImage, ds, pol='xy')
        
    def test_gridding_uvfits(self):
        """Test building a image from a visibility data set - UVFITS file."""
        
        # Open the UVFITS file
        uv = utils.CorrelatedData(uvFile)
        
        # Build the image
        
        ds = uv.getDataSet(1)
        junk = utils.buildGriddedImage(ds, verbose=False)
        
        # Error checking
        self.assertRaises(RuntimeError, utils.buildGriddedImage, ds, pol='xy')
        
    def test_selfcal(self):
        """Test running a simple self calibration."""
        
        # Open the FITS IDI file
        idi = utils.CorrelatedData(idiFile)
        
        # Go for it!
        aa = idi.getAntennaArray()
        ds = idi.getDataSet(1)
        junk = selfCal.phaseOnly(aa, ds, ds, 173, 'xx', nIter=1, verbose=False)
        
        # Error checking
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=0  )
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=564)
        
    def test_selfcal_alt(self):
        """Test running a simple self calibration - alternate FITS IDI file."""
        
        # Open the alternate FITS IDI file
        idi = utils.CorrelatedData(idiAltFile)
        
        # Go for it!
        aa = idi.getAntennaArray()
        ds = idi.getDataSet(1)
        junk = selfCal.phaseOnly(aa, ds, ds, 173, 'xx', nIter=1, verbose=False)
        
        # Error checking
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=0  )
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=564)
        
    def test_selfcal_uvfits(self):
        """Test running a simple self calibration - UVFITS file."""
        
        # Open the alternate UVFITS file
        uv = utils.CorrelatedData(uvFile)
        
        # Go for it!
        aa = uv.getAntennaArray()
        ds = uv.getDataSet(1)
        junk = selfCal.phaseOnly(aa, ds, ds, 173, 'xx', nIter=1, verbose=False)
        
        # Error checking
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=0  )
        self.assertRaises(RuntimeError, selfCal.phaseOnly, aa, ds, ds, 173, 'yx', refAnt=564)
        
    def tearDown(self):
        """Remove the test path directory and its contents"""

        shutil.rmtree(self.testPath, ignore_errors=True)


class imaging_test_suite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the lsl.imaging units 
    tests."""
    
    def __init__(self):
        unittest.TestSuite.__init__(self)
        
        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(imaging_tests)) 


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

"""
Module to make an ASCII progress bar.
"""

import copy
import time

__version__ = '0.2'
__revision__ = '$Rev: 2384 $'
__all__ = ['ProgressBar', 'ProgressBarPlus', '__version__', '__revision__', '__all__']


class ProgressBar(object):
    """
    Object to make a ASCII progress bar for use with various long-
    run tasks.
    
    Example Usage:
        >>> import sys
        >>> from progess import ProgressBar
        >>> pb = ProgressBar()
        >>> pb.inc()
        >>> sys.stdout.write(pb.show())
        >>> sys.stdout.flush()
    """
    
    def __init__(self, max=100, span=70, sym='=', printP=True):
        """
        Initialize the ProgressBar class with various parameters:
        * max: maximum count for the progress bar (default: 100)
        * span: width in characters to make the bar (default: 70)
        * sym: character to use in the progress bar (default: '=')
        * printP: whether or not to print the percentage in addition to
            the bar or not (default: True)
        """
        
        self.amount = 0
        self.max = max
        self.span = span
        self.sym = sym
        self.rotations = ['-', '\\', '|', '/', self.sym]
        self.printP = printP
        
    def inc(self, amount=1):
        """
        Increment the progress bar's internal counter by some amount.  The
        default is one.
        """
        
        self.__iadd__(amount)
        
    def dec(self, amount=1):
        """
        Decrement the progress bar's internal counter by some amount.  The
        default is one.
        """
            
        self.__isub__(amount)
        
    def show(self):
        """
        Build a string representation of the progress bar and return it.
        """
        
        if self.printP:
            # If we want the percentage also displayed, trim a little 
            # more from the progress bar's wdith
            barSpan = self.span - 9
            nMarks = float(self.amount)/self.max * barSpan
            nMarksFull = int(nMarks)
            if nMarksFull < barSpan:
                partial = nMarks - nMarksFull
                lastMark = self.rotations[int(partial*len(self.rotations))]
            else:
                lastMark = ''
            bar = self.sym * nMarksFull
            bar = bar + lastMark
            bar = bar+(' ' * (barSpan-(nMarksFull+len(lastMark))))
            nte = "%5.1f%%" % (float(self.amount)/self.max*100)
            
            out = "[%s] %s" % (bar, nte)
        else:
            # Progress bar only
            barSpan = self.span - 2
            nMarks = float(self.amount)/self.max * barSpan
            nMarksFull = int(nMarks)
            if nMarksFull < barSpan:
                partial = nMarks - nMarksFull
                lastMark = self.rotations[int(partial*len(self.rotations))]
            else:
                lastMark = ''
            bar = self.sym * nMarksFull
            bar = bar + lastMark
            bar = bar+(' ' * (barSpan-(nMarksFull+len(lastMark))))
            
            out = "[%s]" % bar
            
        return out
        
    def __add__(self, amount):
        """
        Increment the internal counter by a certain amount, return a new
        ProgressBar object.
        """
        
        newBar = copy.deepcopy(self)
        newBar += amount
        return newBar
        
    def __iadd__(self, amount):
        """
        Increment the internal counter by a certain amount.
        """
        
        self.amount += amount
        return self
        
    def __sub__(self, amount):
        """
        Decrement the internal counter by a certain amount, return a new
        ProgressBar object.
        """
        
        newBar = copy.deepcopy(self)
        if newBar.amount >= amount:
            newBar -= amount
        return newBar
        
    def __isub__(self, amount):
        """
        Decrement the internal counter by a certain amount.
        """
        
        if self.amount >= amount:
            self.amount -= amount
        return self
        
    def __str__(self):
        """
        Alternative to self.show().
        """
        
        return self.show()


class ProgressBarPlus(ProgressBar):
    """
    Extended version of the ProgressBar class that has a crude time 
    estimator.
    
    Example Usage:
        >>> import sys
        >>> from progess import ProgressBar
        >>> pb = ProgressBar()
        >>> pb.inc()
        >>> sys.stdout.write(pb.show())
        >>> sys.stdout.flush()
        
    .. note::
        The timing feature is only active when the inc()/dec() functions are called.
        
    .. versionadded:: 0.6.4
    """
    
    t0 = None
    t1 = None
    
    def startTimer(self):
        """
        Initialize the timer.
        """
        
        self.t0 = time.time()
        
    def __iadd__(self, amount):
        """
        Increment the internal counter by a certain amount.
        """
        
        self.amount += amount
        
        if self.t0 is None:
            self.t0 = time.time()
        self.t1 = time.time()
        
        return self
        
    def __isub__(self, amount):
        """
        Decrement the internal counter by a certain amount.
        """
        
        if self.amount >= amount:
            self.amount -= amount
            
        if self.t0 is None:
            self.t0 = time.time()
        self.t1 = time.time()
        
        return self
        
    def show(self):
        """
        Build a string representation of the progress bar and return it.
        """
        
        if self.t0 is None:
            # Have we started?
            cte = '----m--s'
        elif self.t1 - self.t0 < 0.2:
            # Have we running long enough to get a "good" estimate?
            cte = '----m--s'
        elif self.amount == 0:
            # Have we gone far enough to get a "good" estimate?
            cte = '----m--s'
        elif self.amount == self.max:
            # Are we done?
            cte = self.t1 - self.t0
            cte = '%4im%02is' % (cte/60, cte%60)
        else:
            cte = (self.max - self.amount) * (self.t1 - self.t0)/self.amount
            cte = '%4im%02is' % (cte/60, cte%60)
            
        if self.printP:
            # If we want the percentage also displayed, trim a little 
            # more from the progress bar's wdith
            barSpan = self.span - 9
            nMarks = float(self.amount)/self.max * barSpan
            nMarksFull = int(nMarks)
            if nMarksFull < barSpan:
                partial = nMarks - nMarksFull
                lastMark = self.rotations[int(partial*len(self.rotations))]
            else:
                lastMark = ''
            bar = self.sym * nMarksFull
            bar = bar + lastMark
            bar = bar+(' ' * (barSpan-(nMarksFull+len(lastMark))))
            nte = "%5.1f%%" % (float(self.amount)/self.max*100)
            
            out = "[%s] %s %s" % (bar, nte, cte)
        else:
            # Progress bar only
            barSpan = self.span - 2
            nMarks = float(self.amount)/self.max * barSpan
            nMarksFull = int(nMarks)
            if nMarksFull < barSpan:
                partial = nMarks - nMarksFull
                lastMark = self.rotations[int(partial*len(self.rotations))]
            else:
                lastMark = ''
            bar = self.sym * nMarksFull
            bar = bar + lastMark
            bar = bar+(' ' * (barSpan-(nMarksFull+len(lastMark))))
            
            out = "[%s] %s" % (bar, cte)
            
        return out